from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class FromdateStep(BaseStep):
    name: Literal['fromdate'] = 'fromdate'
    column: ColumnName
    format: str



from typing import Any, Literal, Union

from pydantic import BaseModel, Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames

from ..conditions import Condition


class IfThenElse(BaseModel):
    class Config(PopulatedWithFieldnames):
        ...

    condition: Condition = Field(alias='if')
    then: Any
    else_value: Union['IfThenElse', Any] = Field(alias='else')


IfThenElse.update_forward_refs()


class IfthenelseStep(BaseStep, IfThenElse):
    class Config(PopulatedWithFieldnames):
        ...

    name: Literal['ifthenelse'] = 'ifthenelse'
    new_column: ColumnName = Field(alias='newColumn')


class IfThenElseStepWithVariables(IfthenelseStep, StepWithVariablesMixin):
    class Config(PopulatedWithFieldnames):
        ...



from typing import TYPE_CHECKING

from pypika import Field, functions
from pypika.dialects import PostgreSQLQuery
from pypika.terms import LiteralValue, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DATE_INFO,
    DataTypeMapping,
    Self,
    SQLTranslator,
    StepContext,
)

if TYPE_CHECKING:
    from pypika.queries import QueryBuilder

    from weaverbird.pipeline.steps import DurationStep


class PostgreSQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.POSTGRES
    QUERY_CLS = PostgreSQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.SIMILAR_TO
    TO_DATE_OP = ToDateOp.TIMESTAMP

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        return LiteralValue(f"{date_column.name} + INTERVAL '{add_date_value} {add_date_unit}'")

    def duration(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "DurationStep",
    ) -> StepContext:
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            functions.Extract(
                step.duration_in, Field(step.end_date_column) - Field(step.start_date_column)
            ).as_(step.new_column_name),
        )
        return StepContext(query, columns + [step.new_column_name])


SQLTranslator.register(PostgreSQLTranslator)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.top import TopStep


def translate_top(step: TopStep) -> List[MongoStep]:
    sort_order = 1 if step.sort == 'asc' else -1
    group_cols = {g: f"${g}" for g in step.groups} if step.groups else None
    return [
        {'$sort': {step.rank_on: sort_order}},
        {'$group': {'_id': group_cols, '_vqbAppArray': {'$push': '$$ROOT'}}},
        {'$project': {'_vqbAppTopElems': {'$slice': ['$_vqbAppArray', step.limit]}}},
        {'$unwind': '$_vqbAppTopElems'},
        {'$replaceRoot': {'newRoot': '$_vqbAppTopElems'}},
    ]



from datetime import datetime
from typing import TYPE_CHECKING, TypeVar

from pypika import Field, Table, functions
from pypika.dialects import MySQLQuery
from pypika.queries import QueryBuilder
from pypika.terms import LiteralValue, Term
from pypika.utils import format_quotes

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DATE_INFO,
    DataTypeMapping,
    SQLTranslator,
    StepContext,
)
from weaverbird.pipeline.steps import UnpivotStep

Self = TypeVar("Self", bound="MySQLTranslator")

if TYPE_CHECKING:
    from weaverbird.pipeline.steps import SplitStep


class MySQLTranslator(SQLTranslator):
    DIALECT = SQLDialect.MYSQL
    QUERY_CLS = MySQLQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="SIGNED",
        date="DATE",
        float="DECIMAL",
        integer="UNSIGNED",
        text="CHAR",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    # Requires MySQL>=8
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = False
    FROM_DATE_OP = FromDateOp.DATE_FORMAT
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.STR_TO_DATE

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        return LiteralValue(
            f"DATE_ADD({date_column.name}, INTERVAL {add_date_value} {add_date_unit})"
        )

    @classmethod
    def _build_unpivot_col(
        cls, *, step: 'UnpivotStep', quote_char: str | None, secondary_quote_char: str
    ) -> str:
        value_col = format_quotes(step.value_column_name, quote_char)
        unpivot_col = format_quotes(step.unpivot_column_name, quote_char)
        in_cols = [format_quotes(col, quote_char) for col in step.unpivot]
        in_single_quote_cols = [format_quotes(col, secondary_quote_char) for col in step.unpivot]
        rows = ' UNION ALL SELECT '.join(
            [f'{col}, {val}' for col, val in zip(in_cols, in_single_quote_cols)]
        )
        return f' t1 CROSS JOIN LATERAL(SELECT {rows}) AS t2({value_col}, {unpivot_col})'

    def split(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SplitStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]
        new_cols = [f"{step.column}_{i + 1}" for i in range(step.number_cols_to_keep)]

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            *(
                # https://stackoverflow.com/a/32500349
                SubstringIndex(
                    SubstringIndex(col_field, step.delimiter, i + 1), step.delimiter, -1
                ).as_(new_cols[i])
                for i in range(step.number_cols_to_keep)
            ),
        )
        return StepContext(query, columns + new_cols)

    @staticmethod
    def _cast_to_timestamp(value: str | datetime | Field | Term) -> functions.Function:
        return functions.Timestamp(value)


SQLTranslator.register(MySQLTranslator)


class SubstringIndex(functions.Function):  # type: ignore[misc]
    def __init__(
        self, term: str | Field, delimiter: str, count: int, alias: str | None = None
    ) -> None:
        super().__init__("SUBSTRING_INDEX", term, delimiter, count, alias=alias)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import AbsoluteValueStep


def translate_absolutevalue(step: AbsoluteValueStep) -> List[MongoStep]:
    return [{'$addFields': {step.new_column: {'$abs': f'${step.column}'}}}]



from typing import Literal, Optional

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class ToDateStep(BaseStep):
    name: Literal['todate'] = 'todate'
    column: ColumnName
    format: Optional[str]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DuplicateStep


def translate_duplicate(step: DuplicateStep) -> List[MongoStep]:
    return [
        {
            '$addFields': {(step.new_column_name): f'${step.column}'},
        }
    ]



from typing import Dict, List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import RollupStep


def column_map(s: List[str]) -> Dict[str, str]:
    return {e: f'${e}' for e in s}


def translate_rollup(step: RollupStep) -> List[MongoStep]:
    facet: Dict[str, List[MongoStep]] = {}
    label_col = step.label_col or 'label'
    level_col = step.level_col or 'level'
    parent_label_col = step.parent_label_col or 'parent'
    add_fields = {}

    for idx, elem in enumerate(step.hierarchy):
        id = column_map(step.hierarchy[: idx + 1] + (step.groupby or []))
        aggs: Dict[str, dict] = {}
        for agg_step in step.aggregations:
            cols = agg_step.columns
            new_cols = agg_step.new_columns

            if agg_step.agg_function == 'count':
                for i in range(len(cols)):
                    # cols and new_cols are always of same length
                    aggs[new_cols[i]] = {'$sum': 1}
            elif agg_step.agg_function == 'count distinct':
                # specific step needed to count distinct values
                for i in range(len(cols)):
                    # build a set of unique values
                    aggs[new_cols[i]] = {'$addToSet': f'${cols[i]}'}
                    # count the number of items in the set
                    add_fields[new_cols[i]] = {'$size': f'${new_cols[i]}'}
            else:
                for i in range(len(cols)):
                    # cols and new_cols are always of same length
                    aggs[new_cols[i]] = {f'${agg_step.agg_function}': f'${cols[i]}'}

        project: dict = {
            '_id': 0,
            **{k: f'$_id.{k}' for k in id.keys()},
            **{k: 1 for k in aggs.keys()},
            label_col: f'$_id.{elem}',
            level_col: elem,
        }

        if idx > 0:
            project[parent_label_col] = f'$_id.{step.hierarchy[idx - 1]}'

        addFieldsToAddToPipeline = [{'$addFields': add_fields}] if add_fields else []

        facet[f'level_{idx}'] = [
            {
                '$group': {
                    '_id': id,
                    **aggs,
                },
            },
            *addFieldsToAddToPipeline,
            {
                '$project': project,
            },
        ]

    return [
        {'$facet': facet},
        {
            '$project': {
                '_vqbRollupLevels': {
                    '$concatArrays': [f'${k}' for k in sorted(facet.keys())],
                },
            },
        },
        {'$unwind': '$_vqbRollupLevels'},
        {'$replaceRoot': {'newRoot': '$_vqbRollupLevels'}},
    ]



from datetime import datetime
from typing import Literal, Optional, Union

from pydantic import BaseModel


class RelativeDate(BaseModel):
    date: Optional[Union[datetime, str]]
    operator: Literal['from', 'until', 'before', 'after']
    quantity: int
    duration: Literal['year', 'quarter', 'month', 'week', 'day']



from abc import ABC
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import cache
from typing import TYPE_CHECKING, Any, Callable, Literal, Mapping, Sequence, TypeVar, Union, cast

from dateutil import parser as dateutil_parser
from pypika import (
    AliasedQuery,
    Case,
    Criterion,
    Field,
    Order,
    Query,
    Schema,
    Table,
    Tables,
    analytics,
    functions,
)
from pypika.enums import Comparator, JoinType
from pypika.queries import QueryBuilder, Selectable
from pypika.terms import AnalyticFunction, BasicCriterion, LiteralValue, Term
from pypika.utils import format_quotes

from weaverbird.backends.pandas_executor.steps.utils.dates import evaluate_relative_date
from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators import ALL_TRANSLATORS
from weaverbird.backends.pypika_translator.utils.formula import formula_to_term
from weaverbird.pipeline.conditions import (
    ComparisonCondition,
    DateBoundCondition,
    InclusionCondition,
    MatchCondition,
    NullCondition,
)
from weaverbird.pipeline.dates import RelativeDate
from weaverbird.pipeline.pipeline import Pipeline
from weaverbird.pipeline.steps.date_extract import DATE_INFO
from weaverbird.pipeline.steps.utils.combination import PipelineOrDomainNameOrReference, Reference

Self = TypeVar("Self", bound="SQLTranslator")

if TYPE_CHECKING:
    from weaverbird.pipeline import PipelineStep
    from weaverbird.pipeline.conditions import Condition, SimpleCondition
    from weaverbird.pipeline.steps import (
        AbsoluteValueStep,
        AggregateStep,
        AppendStep,
        ArgmaxStep,
        ArgminStep,
        CompareTextStep,
        ConcatenateStep,
        ConvertStep,
        CustomSqlStep,
        DateExtractStep,
        DeleteStep,
        DomainStep,
        DuplicateStep,
        EvolutionStep,
        FillnaStep,
        FilterStep,
        FormulaStep,
        FromdateStep,
        IfthenelseStep,
        JoinStep,
        LowercaseStep,
        PercentageStep,
        RenameStep,
        ReplaceStep,
        SelectStep,
        SortStep,
        SplitStep,
        SubstringStep,
        TextStep,
        ToDateStep,
        TopStep,
        TrimStep,
        UniqueGroupsStep,
        UnpivotStep,
        UppercaseStep,
    )
    from weaverbird.pipeline.steps.aggregate import AggregateFn
    from weaverbird.pipeline.steps.evolution import EVOLUTION_TYPE


@dataclass(kw_only=True)
class DataTypeMapping:
    boolean: str
    date: str
    float: str
    integer: str
    text: str
    datetime: str
    timestamp: str


@dataclass
class StepContext:
    selectable: Selectable | LiteralValue
    columns: list[str]
    builder: QueryBuilder | None = None

    def update_builder(self, builder: QueryBuilder, step_name: str) -> QueryBuilder:
        builder = builder if self.builder is None else self.builder
        return builder.with_(self.selectable, step_name)


@dataclass(kw_only=True)
class QueryBuilderContext:
    builder: QueryBuilder
    columns: list[str]
    table_name: str

    def materialize(self) -> QueryBuilder:
        return self.builder.from_(self.table_name).select(*self.columns)


class CustomQuery(AliasedQuery):  # type: ignore[misc]
    def get_sql(self, **kwargs: Any) -> str:
        return cast(str, self.query)


class SQLTranslator(ABC):
    DIALECT: SQLDialect
    QUERY_CLS: Query
    DATA_TYPE_MAPPING: DataTypeMapping
    # supported extra functions
    SUPPORT_ROW_NUMBER: bool
    SUPPORT_SPLIT_PART: bool
    SUPPORT_UNPIVOT: bool = False
    # which operators should be used
    FROM_DATE_OP: FromDateOp
    REGEXP_OP: RegexOp
    TO_DATE_OP: ToDateOp
    EVOLUTION_DATE_UNIT: dict['EVOLUTION_TYPE', DATE_INFO] = {
        'vsLastYear': 'year',
        'vsLastMonth': 'month',
        'vsLastWeek': 'week',
        'vsLastDay': 'day',
    }

    def __init__(
        self: Self,
        *,
        tables_columns: Mapping[str, Sequence[str]] | None = None,
        db_schema: str | None = None,
        known_instances: dict[int, str] | None = None,
    ) -> None:
        self._tables_columns: Mapping[str, Sequence[str]] = tables_columns or {}
        self._db_schema_name = db_schema
        self._db_schema: Schema | None = Schema(db_schema) if db_schema is not None else None
        self._known_instances: dict[int, str] = known_instances or {}
        self._step_count = 0

    def __init_subclass__(cls) -> None:
        ALL_TRANSLATORS[cls.DIALECT] = cls

    @cache
    def _id(self: Self) -> str:
        if id(self) in self._known_instances:
            return self._known_instances[id(self)]
        if len(self._known_instances.keys()) == 0:
            id_ = self._known_instances[id(self)] = self.__class__.__name__.lower()
            return id_
        else:
            id_ = self.__class__.__name__.lower() + str(len(self._known_instances.keys()))
            self._known_instances[id(self)] = id_
            return id_

    def _step_name(self: Self) -> str:
        return f'__step_{self._step_count}_{self._id()}__'

    def _next_step_name(self: Self) -> str:
        name = self._step_name()
        self._step_count += 1
        return name

    def _step_context_from_first_step(
        self, step: Union['DomainStep', 'CustomSqlStep']
    ) -> StepContext:
        return (
            self._domain(step=step)
            if step.name == 'domain'
            else StepContext(self._custom_query(step=step), ['*'])
        )

    def get_query_builder(
        self: Self,
        *,
        steps: Sequence["PipelineStep"],
        query_builder: QueryBuilder | None = None,
    ) -> QueryBuilderContext:
        if len(steps) < 0:
            ValueError('No steps provided')
        assert steps[0].name == "domain" or steps[0].name == "customsql"
        self._step_count = 0

        ctx = self._step_context_from_first_step(steps[0])
        table_name = self._next_step_name()
        builder = (query_builder if query_builder is not None else self.QUERY_CLS).with_(
            ctx.selectable, table_name
        )

        for step in steps[1:]:
            step_method: Callable[..., StepContext] | None = getattr(self, step.name, None)
            if step_method is None:
                raise NotImplementedError(f"[{self.DIALECT}] step {step.name} is not implemented")
            ctx = step_method(
                step=step, prev_step_name=table_name, builder=builder, columns=ctx.columns
            )
            table_name = self._next_step_name()
            builder = ctx.update_builder(builder=builder, step_name=table_name)
        return QueryBuilderContext(builder=builder, columns=ctx.columns, table_name=table_name)

    def get_query_str(self: Self, *, steps: Sequence["PipelineStep"]) -> str:
        return self.get_query_builder(steps=steps).materialize().get_sql()

    # All other methods implement step from https://weaverbird.toucantoco.com/docs/steps/,
    # the name of the method being the name of the step and the kwargs the rest of the params
    def _get_aggregate_function(
        self: Self, agg_function: "AggregateFn"
    ) -> type[functions.AggregateFunction] | None:
        match agg_function:
            case "avg":
                return functions.Avg
            case "count":
                return functions.Count
            case "count distinct":
                return CountDistinct
            case "max":
                return functions.Max
            case "min":
                return functions.Min
            case "sum":
                return functions.Sum
            case _:
                return None

    def _get_window_function(
        self: Self, window_function: "AggregateFn"
    ) -> analytics.AnalyticFunction | None:
        match window_function:
            case 'first':
                return analytics.FirstValue
            case 'last':
                return analytics.LastValue
            case _:
                return None

    def absolutevalue(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "AbsoluteValueStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]
        query: "Selectable" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, functions.Abs(col_field).as_(step.new_column)
        )
        return StepContext(query, columns + [step.new_column])

    def aggregate(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "AggregateStep",
    ) -> StepContext:
        agg_selected: list[Field] = []
        window_selected: list[tuple[int, Field]] = []
        window_subquery_list: list[Tables] = []

        def _build_window_subquery() -> Any:
            min_window_index = min(c[0] for c in window_selected)
            first_wq = Table(f'wq{min_window_index}')
            merged_query = (
                self.QUERY_CLS.from_(window_subquery_list[0])
                .select(
                    *step.on,
                    *[
                        getattr(first_wq, col[1].alias)
                        for col in window_selected
                        if col[0] == min_window_index
                    ],
                )
                .as_('window_subquery')
            )
            for index, sq in enumerate(window_subquery_list[1:]):
                wq_temp = Table(f'wq{min_window_index + index + 1}')
                merged_query = (
                    merged_query.join(sq)
                    .on_field(*step.on)
                    .select(
                        *[
                            getattr(wq_temp, col[1].alias)
                            for col in window_selected
                            if col[0] == min_window_index + index + 1
                        ]
                    )
                    .as_('window_subquery')
                )
            return merged_query

        # Handle aggregation and analytics functions in distinct subqueries

        for step_index, aggregation in enumerate(step.aggregations):
            if agg_fn := self._get_aggregate_function(aggregation.agg_function):
                for agg_column_name, new_column_name in zip(
                    aggregation.columns, aggregation.new_columns
                ):
                    column_field: Field = Table(prev_step_name)[agg_column_name]
                    new_agg_col = agg_fn(column_field).as_(new_column_name)
                    agg_selected.append(new_agg_col)

            elif window_fn := self._get_window_function(aggregation.agg_function):
                agg_cols: list[Field] = []
                for window_index, window_column_name in enumerate(aggregation.columns):
                    column_field = Table(prev_step_name)[window_column_name]
                    new_window_col = (
                        window_fn(column_field)
                        .over(*step.on)
                        .orderby(column_field)
                        .rows(analytics.Preceding(), analytics.Following())
                        .as_(aggregation.new_columns[window_index])
                    )

                    window_selected.append((step_index, new_window_col))
                    agg_cols.append(new_window_col)
                window_subquery_list.append(
                    self.QUERY_CLS.from_(prev_step_name)
                    .select(*step.on, *agg_cols)
                    .distinct()
                    .as_(f'wq{step_index}')
                )

            else:  # pragma: no cover
                raise NotImplementedError(
                    f"[{self.DIALECT}] Aggregation for {aggregation.agg_function!r} is not yet implemented"
                )
        if window_subquery_list and agg_selected:
            window_table = Table('window_subquery')
            all_windows_subquery = _build_window_subquery()
            agg_query = (
                self.QUERY_CLS.from_(prev_step_name)
                .select(*agg_selected, *step.on)
                .groupby(*step.on)
                .orderby(*step.on, order=Order.asc)
            ).as_('agg_subquery')
            agg_table = Table('agg_subquery')
            merged_selected: list[str | Field] = [
                *step.on,
                *[getattr(agg_table, col.alias) for col in agg_selected],
                *[getattr(window_table, col[1].alias) for col in window_selected],
            ]
            merged_query = (
                self.QUERY_CLS.from_(agg_query)
                .select(*merged_selected)
                .inner_join(all_windows_subquery)
                .on_field(*step.on)
            )
        elif agg_selected:
            selected_cols = [*step.on, *agg_selected]
            merged_query = (
                self.QUERY_CLS.from_(prev_step_name).select(*selected_cols).groupby(*step.on)
            )
        elif window_subquery_list:
            merged_query = _build_window_subquery()
        else:
            merged_query = self.QUERY_CLS.from_(prev_step_name).groupby(*step.on).select(*step.on)
        query: "QueryBuilder"
        selected_col_names: list[str]

        if step.keep_original_granularity:
            all_agg_col_names: list[str] = [x for agg in step.aggregations for x in agg.new_columns]
            query = (
                self.QUERY_CLS.from_(prev_step_name)
                .select(
                    *columns, *(Field(agg_col, table=merged_query) for agg_col in all_agg_col_names)
                )
                .left_join(merged_query)
                .on_field(*step.on)
            )
            selected_col_names = [*columns, *all_agg_col_names]
            return StepContext(query.orderby(*step.on) if step.on else query, selected_col_names)

        else:
            selected_col_names = [
                *step.on,
                *(f.alias for f in agg_selected),
                *(f[1].alias for f in window_selected),
            ]
            return StepContext(
                merged_query.orderby(*step.on) if step.on else merged_query,
                selected_col_names,
            )

    @staticmethod
    def _pipeline_or_domain_name_or_reference_to_pipeline(
        pipeline: 'PipelineOrDomainNameOrReference', message: str | None = None
    ) -> list['PipelineStep']:
        try:
            return Pipeline(steps=pipeline).steps
        except Exception as exc:
            message = message or f"could not convert {pipeline} to pipeline: {exc}"
            raise NotImplementedError(message) from exc

    def append(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "AppendStep",
    ) -> StepContext:
        pipelines = [
            self._pipeline_or_domain_name_or_reference_to_pipeline(pipeline)
            for pipeline in step.pipelines
        ]
        tables: list[str] = []
        column_lists: list[list[str]] = []
        for pipeline in pipelines:
            pipeline_ctx = self.__class__(
                tables_columns=self._tables_columns,
                db_schema=self._db_schema_name,
                known_instances=self._known_instances,
            ).get_query_builder(steps=pipeline, query_builder=builder)
            tables.append(pipeline_ctx.table_name)
            column_lists.append(pipeline_ctx.columns)
            builder = pipeline_ctx.builder

        # By default, UNION ALL will append columns by index in SQL, and only append as much columns
        # as available in the first SELECT statement. Also, the columns are appended in the order of
        # the select statement, which is not what we want either (a merge by name is probably what's
        # expected by the user)
        columns_to_add = []
        for column_list in column_lists:
            for col in column_list:
                if col not in columns and col not in columns_to_add:
                    columns_to_add.append(col)

        all_columns = columns + columns_to_add

        query = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, *(LiteralValue('NULL').as_(col) for col in columns_to_add)
        )
        for table, column_list in zip(tables, column_lists):
            query = query.union_all(
                self.QUERY_CLS.from_(table).select(
                    *(
                        # Selecting either the column from the dataset if it is available, or NULL
                        # AS "col" otherwise. We iterate over all_columns rather than column_list in
                        # order to have a merge by name
                        col if col in column_list else LiteralValue('NULL').as_(col)
                        for col in all_columns
                    )
                )
            )

        return StepContext(query.orderby(*all_columns), all_columns, builder)

    def argmax(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ArgmaxStep",
    ) -> StepContext:
        from weaverbird.pipeline.steps import TopStep

        return self.top(
            builder=builder,
            prev_step_name=prev_step_name,
            columns=columns,
            step=TopStep(rank_on=step.column, sort="desc", limit=1, groups=step.groups),
        )

    def argmin(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ArgminStep",
    ) -> StepContext:
        from weaverbird.pipeline.steps import TopStep

        return self.top(
            builder=builder,
            prev_step_name=prev_step_name,
            columns=columns,
            step=TopStep(rank_on=step.column, sort="asc", limit=1, groups=step.groups),
        )

    def comparetext(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "CompareTextStep",
    ) -> StepContext:
        table = Table(prev_step_name)
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            Case()
            .when(table[step.str_col_1] == table[step.str_col_2], True)
            .else_(False)
            .as_(step.new_column_name),
        )
        return StepContext(query, columns + [step.new_column_name])

    def concatenate(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ConcatenateStep",
    ) -> StepContext:
        # from step.columns = ["city", "age", "username"], step.separator = " -> "
        # create [Field("city"), " -> ", Field("age"), " -> ", Field("username")]
        the_table = Table(prev_step_name)
        tokens = [the_table[step.columns[0]]]
        for col in step.columns[1:]:
            tokens.append(step.separator)
            tokens.append(the_table[col])

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            functions.Concat(*tokens).as_(step.new_column_name),
        )
        return StepContext(query, columns + [step.new_column_name])

    def convert(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ConvertStep",
    ) -> StepContext:
        col_fields: list[Field] = [Table(prev_step_name)[col] for col in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c not in step.columns),
            *(
                functions.Cast(col_field, getattr(self.DATA_TYPE_MAPPING, step.data_type)).as_(
                    col_field.name
                )
                for col_field in col_fields
            ),
        )
        return StepContext(query, columns)

    def _custom_query(
        self: Self, *, step: 'CustomSqlStep', prev_step_name: str | None = None
    ) -> CustomQuery:
        table_name = prev_step_name or '_'
        return CustomQuery(
            name=f"custom_from_{table_name}",
            query=step.query.replace("##PREVIOUS_STEP##", table_name),
        )

    def customsql(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "CustomSqlStep",
    ) -> StepContext:
        """create a custom sql step based on the current table named ##PREVIOUS_STEP## in the query"""
        # we have no way to know which columns remain without actually executing the query
        return StepContext(self._custom_query(step=step), ["*"])

    @classmethod
    def _get_date_extract_func(cls, *, date_unit: DATE_INFO, target_column: Field) -> LiteralValue:
        # TODO to implement for other connectors than Snowflake
        raise NotImplementedError(f"[{cls.DIALECT}] _get_date_extract_func is not implemented")

    def dateextract(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "DateExtractStep",
    ) -> StepContext:
        date_col: Field = Table(prev_step_name)[step.column]
        extracted_dates: list[LiteralValue] = []

        for date_info, new_column_name in zip(step.date_info, step.new_columns):
            extracted_dates.append(
                self._get_date_extract_func(date_unit=date_info, target_column=date_col).as_(
                    new_column_name
                )
            )
        query: "Selectable" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, *extracted_dates
        )

        return StepContext(query, columns + [col.alias for col in extracted_dates])

    def delete(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "DeleteStep",
    ) -> StepContext:
        new_columns = [c for c in columns if c not in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(*new_columns)
        return StepContext(query, new_columns)

    # Prefixing domain with a '_', as it is a special case and should not be returned by
    # getattr(self, step_name)
    def _domain(self: Self, *, step: "DomainStep") -> StepContext:
        try:
            if isinstance(step.domain, Reference):
                raise NotImplementedError(f"[{self.DIALECT}] Cannot resolve a reference to a query")
            else:
                selected_cols = list(self._tables_columns[step.domain])
        except KeyError:
            selected_cols = ["*"]

        query: "QueryBuilder" = self.QUERY_CLS.from_(
            Table(step.domain, schema=self._db_schema)
        ).select(*selected_cols)
        return StepContext(query, selected_cols)

    def duplicate(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "DuplicateStep",
    ) -> StepContext:
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, Table(prev_step_name)[step.column].as_(step.new_column_name)
        )
        return StepContext(query, columns + [step.new_column_name])

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        raise NotImplementedError(f"[{cls.DIALECT}] _add_date is not implemented")

    def evolution(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "EvolutionStep",
    ) -> StepContext:

        prev_table = Table(prev_step_name)
        lagged_date = self._add_date(
            date_column=prev_table.field(step.date_col),
            add_date_value=1,
            add_date_unit=self.EVOLUTION_DATE_UNIT[step.evolution_type],
        ).as_(step.date_col)
        right_table = Table('right_table')
        new_col = step.new_column if step.new_column else 'evol'
        query: "QueryBuilder" = (
            self.QUERY_CLS.from_(prev_step_name)
            .select(
                *[prev_table.field(col) for col in columns],
                (
                    prev_table.field(step.value_col) - right_table.field(step.value_col)
                    if step.evolution_format == 'abs'
                    else prev_table.field(step.value_col) / right_table.field(step.value_col) - 1
                ).as_(new_col),
                *[prev_table.field(col).as_(f'left_table_{col}') for col in step.index_columns],
            )
            .left_join(
                self.QUERY_CLS.from_(prev_step_name)
                .select(
                    step.value_col,
                    lagged_date,
                    *step.index_columns,
                )
                .as_('right_table'),
            )
            .on_field(step.date_col, *step.index_columns)
            .orderby(step.date_col)
        )
        return StepContext(query, columns + [new_col])

    def fillna(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "FillnaStep",
    ) -> StepContext:
        the_table = Table(prev_step_name)
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c not in step.columns),
            *(
                functions.Coalesce(the_table[col_name], step.value).as_(col_name)
                for col_name in step.columns
            ),
        )
        return StepContext(query, columns)

    @staticmethod
    def _cast_to_timestamp(value: str | datetime | Field | Term) -> functions.Function:
        return functions.Cast(value, "TIMESTAMP")

    def _get_single_condition_criterion(
        self: Self, condition: "SimpleCondition", prev_step_name: str
    ) -> Criterion:
        column_field: Field = Table(prev_step_name)[condition.column]

        match condition:
            case ComparisonCondition():
                import operator

                op = getattr(operator, condition.operator)
                return op(column_field, condition.value)

            case InclusionCondition():
                if condition.operator == "in":
                    return column_field.isin(condition.value)
                elif condition.operator == "nin":
                    return column_field.notin(condition.value)

            case MatchCondition():
                compliant_regex = _compliant_regex(condition.value, self.DIALECT)

                if condition.operator == "matches":
                    match self.REGEXP_OP:
                        case RegexOp.REGEXP:
                            return BasicCriterion(
                                RegexpMatching.regexp,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.SIMILAR_TO:
                            return BasicCriterion(
                                RegexpMatching.similar_to,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.CONTAINS:
                            return BasicCriterion(
                                RegexpMatching.contains,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_CONTAINS:
                            return functions.Function(
                                RegexOp.REGEXP_CONTAINS,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_LIKE:
                            return functions.Function(
                                RegexOp.REGEXP_LIKE,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_CONTAINS:
                            return functions.Function(
                                RegexOp.NOT_REGEXP_CONTAINS,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_LIKE:
                            return functions.Function(
                                RegexOp.NOT_REGEXP_LIKE,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case _:
                            raise NotImplementedError(
                                f"[{self.DIALECT}] doesn't have regexp operator"
                            )

                elif condition.operator == "notmatches":
                    match self.REGEXP_OP:
                        case RegexOp.REGEXP:
                            return column_field.regexp(condition.value).negate()
                        case RegexOp.SIMILAR_TO:
                            return BasicCriterion(
                                RegexpMatching.not_similar_to,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.CONTAINS:
                            return BasicCriterion(
                                RegexpMatching.not_contains,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_CONTAINS:
                            return functions.Function(
                                RegexOp.NOT_REGEXP_CONTAINS,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case RegexOp.REGEXP_LIKE:
                            return functions.Function(
                                RegexOp.NOT_REGEXP_LIKE,
                                column_field,
                                column_field.wrap_constant(compliant_regex),
                            )
                        case _:
                            raise NotImplementedError(
                                f"[{self.DIALECT}] doesn't have regexp operator"
                            )

            case NullCondition():
                if condition.operator == "isnull":
                    return column_field.isnull()
                elif condition.operator == "notnull":
                    return column_field.isnotnull()

            case DateBoundCondition():

                if isinstance(condition.value, (RelativeDate, datetime, str)):
                    if isinstance(condition.value, RelativeDate):
                        dt = evaluate_relative_date(condition.value)
                    elif isinstance(condition.value, datetime):
                        dt = condition.value
                    else:
                        dt = dateutil_parser.parse(condition.value)
                    dt = (
                        dt.replace(tzinfo=timezone.utc)
                        if dt.tzinfo is None
                        else dt.astimezone(timezone.utc)
                    )
                    value_to_compare = self._cast_to_timestamp(dt.strftime('%Y-%m-%d %H:%M:%S'))

                elif isinstance(condition.value, functions.Function):
                    value_to_compare = condition.value

                if condition.operator == "from":
                    return self._cast_to_timestamp(column_field) >= value_to_compare
                elif condition.operator == "until":
                    return self._cast_to_timestamp(column_field) <= value_to_compare

            case _:  # pragma: no cover
                raise KeyError(f"Operator {condition.operator!r} does not exist")

    def _get_filter_criterion(self: Self, condition: "Condition", prev_step_name: str) -> Criterion:
        from weaverbird.pipeline.conditions import ConditionComboAnd, ConditionComboOr

        match condition:
            case ConditionComboOr():
                return Criterion.any(
                    (
                        self._get_filter_criterion(condition, prev_step_name)
                        for condition in condition.or_
                    )
                )
            case ConditionComboAnd():
                return Criterion.all(
                    (
                        self._get_filter_criterion(condition, prev_step_name)
                        for condition in condition.and_
                    )
                )
            case _:
                return self._get_single_condition_criterion(condition, prev_step_name)

    def filter(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "FilterStep",
    ) -> StepContext:

        query: "QueryBuilder" = (
            self.QUERY_CLS.from_(prev_step_name)
            .select(*columns)
            .where(self._get_filter_criterion(step.condition, prev_step_name))
        )
        return StepContext(query, columns)

    def formula(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "FormulaStep",
    ) -> StepContext:
        formula = formula_to_term(step.formula, Table(prev_step_name))
        query = Query.from_(prev_step_name).select(*(columns), formula.as_(step.new_column))
        return StepContext(query, columns + [step.new_column])

    def fromdate(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "FromdateStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]

        match self.FROM_DATE_OP:
            case FromDateOp.DATE_FORMAT:
                convert_fn = DateFormat
            case FromDateOp.TO_CHAR:
                convert_fn = functions.ToChar
            case _:
                raise NotImplementedError(f"[{self.DIALECT}] doesn't have from date operator")

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.column),
            convert_fn(col_field, step.format).as_(step.column),
        )
        return StepContext(query, columns)

    def _build_ifthenelse_case(
        self,
        *,
        if_: "Condition",
        then_: Any,
        else_: "Condition" | Any,
        prev_step_name: str,
        case_: Case,
    ) -> Case:
        import json

        from weaverbird.pipeline.steps.ifthenelse import IfThenElse

        try:
            # if the value is a string
            then_value = json.loads(then_)
            case_ = case_.when(self._get_filter_criterion(if_, prev_step_name), then_value)
        except (json.JSONDecodeError, TypeError):
            # the value is a formula
            then_value = then_
            case_ = case_.when(
                self._get_filter_criterion(if_, prev_step_name), LiteralValue(then_value)
            )

        if isinstance(else_, IfThenElse):
            return self._build_ifthenelse_case(
                if_=else_.condition,
                then_=else_.then,
                else_=else_.else_value,
                prev_step_name=prev_step_name,
                case_=case_,
            )
        else:
            try:
                # the value is a string
                else_value = json.loads(else_)  # type: ignore
                return case_.else_(else_value)
            except (json.JSONDecodeError, TypeError):
                # the value is a formula
                else_value = else_
                return case_.else_(LiteralValue(else_value))

    def ifthenelse(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "IfthenelseStep",
    ) -> StepContext:
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            self._build_ifthenelse_case(
                if_=step.condition,
                then_=step.then,
                else_=step.else_value,
                prev_step_name=prev_step_name,
                case_=Case(),
            ).as_(step.new_column),
        )
        return StepContext(query, columns + [step.new_column])

    @staticmethod
    def _get_join_type(join_type: Literal['left', 'inner', 'left outer']) -> JoinType:
        match join_type:
            case 'left':
                return JoinType.left
            case 'left outer':
                return JoinType.left_outer
            case 'inner':
                return JoinType.inner

    @staticmethod
    def _field_list_to_name_list(fields: list[Field]) -> list[str]:
        return [field.alias or field.name for field in fields]

    def join(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "JoinStep",
    ) -> StepContext:
        steps = self._pipeline_or_domain_name_or_reference_to_pipeline(step.right_pipeline)

        right_builder_ctx = self.__class__(
            tables_columns=self._tables_columns,
            db_schema=self._db_schema_name,
            known_instances=self._known_instances,
        ).get_query_builder(steps=steps, query_builder=builder)
        left_table = Table(prev_step_name)
        right_table = Table(right_builder_ctx.table_name)

        left_cols = [Field(col, table=left_table) for col in columns]
        # Simply checking if a column is in left columns and right columns is not enough to
        # determine if it needs to be aliased. In case of nested joins where a column is used in
        # three or more datasets (not unlikely with columns like "id" or "name"), the aliased column
        # may need to be suffixed several times. The join/inner_pypika_nested.yaml fixture shows an
        # example of this kind of situation
        right_cols: list[Field] = []
        for col in right_builder_ctx.columns:
            if col not in (all_cols := self._field_list_to_name_list(left_cols + right_cols)):
                right_cols.append(Field(col, table=right_table))
                continue
            alias = col
            while (alias := f'{alias}_right') in all_cols:
                pass
            right_cols.append(Field(col, table=right_table, alias=alias))

        query = (
            self.QUERY_CLS.from_(left_table)
            .select(*left_cols, *right_cols)
            .join(right_table, self._get_join_type(step.type))
            .on(
                Criterion.all(
                    Field(f[0], table=left_table) == Field(f[1], table=right_table) for f in step.on
                )
            )
            # Order of results is not consistent depending on the SQL Engine (inconsistencies
            # observed with Athena and BigQuery).
            .orderby(*(c[0] for c in step.on))
        )
        return StepContext(
            query,
            [c.name for c in left_cols] + [c.alias or c.name for c in right_cols],
            right_builder_ctx.builder,
        )

    def lowercase(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "LowercaseStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.column),
            functions.Lower(col_field).as_(step.column),
        )
        return StepContext(query, columns)

    def percentage(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "PercentageStep",
    ) -> StepContext:
        raise NotImplementedError(f"[{self.DIALECT}] percentage is not implemented")

    def rename(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "RenameStep",
    ) -> StepContext:
        new_names_mapping: dict[str, str] = dict(step.to_rename)

        selected_col_fields: list[Field] = []

        for col_name in columns:
            if col_name in new_names_mapping:
                selected_col_fields.append(Field(name=col_name, alias=new_names_mapping[col_name]))
            else:
                selected_col_fields.append(Field(name=col_name))

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(*selected_col_fields)
        return StepContext(query, [f.alias or f.name for f in selected_col_fields])

    def replace(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ReplaceStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.search_column]

        # Do a nested `replace` to replace many values on the same column
        replaced_col = col_field
        for old_name, new_name in step.to_replace:
            replaced_col = functions.Replace(replaced_col, old_name, new_name)

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.search_column),
            replaced_col.as_(step.search_column),
        )

        return StepContext(query, columns)

    def select(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SelectStep",
    ) -> StepContext:
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(*step.columns)
        return StepContext(query, step.columns)

    def sort(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SortStep",
    ) -> StepContext:
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(*columns)

        for column_sort in step.columns:
            query = query.orderby(
                column_sort.column, order=Order.desc if column_sort.order == "desc" else Order.asc
            )

        return StepContext(query, columns)

    def split(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SplitStep",
    ) -> StepContext:
        if self.SUPPORT_SPLIT_PART:
            col_field: Field = Table(prev_step_name)[step.column]
            new_cols = [f"{step.column}_{i + 1}" for i in range(step.number_cols_to_keep)]
            query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
                *columns,
                *(
                    functions.SplitPart(col_field, step.delimiter, i + 1).as_(new_cols[i])
                    for i in range(step.number_cols_to_keep)
                ),
            )
            return StepContext(query, columns + new_cols)

        raise NotImplementedError(f"[{self.DIALECT}] split is not implemented")

    def substring(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SubstringStep",
    ) -> StepContext:
        step.new_column_name = (
            f"{step.column}_substr" if step.new_column_name is None else step.new_column_name
        )
        col_field: Field = Table(prev_step_name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns,
            functions.Substring(
                col_field, step.start_index, (step.end_index - step.start_index) + 1
            ).as_(step.new_column_name),
        )
        return StepContext(query, columns + [step.new_column_name])

    def text(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "TextStep",
    ) -> StepContext:
        from pypika.terms import ValueWrapper

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, ValueWrapper(step.text).as_(step.new_column)
        )
        return StepContext(query, columns + [step.new_column])

    def todate(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "ToDateStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]

        if step.format is not None:
            match self.TO_DATE_OP:
                case ToDateOp.STR_TO_DATE:
                    convert_fn = StrToDate
                case ToDateOp.PARSE_DATE:
                    convert_fn = ParseDate
                case ToDateOp.TIMESTAMP:
                    convert_fn = functions.Timestamp
                case ToDateOp.TO_TIMESTAMP_NTZ:
                    convert_fn = ToTimestampNTZ
                case _:
                    raise NotImplementedError(f"[{self.DIALECT}] todate has no set operator")
            date_selection = (
                self._cast_to_timestamp(convert_fn(col_field, step.format))
                if convert_fn != ToDateOp.TIMESTAMP
                else convert_fn(col_field, step.format)
            )
        else:
            date_selection = self._cast_to_timestamp(col_field)

        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.column),
            date_selection.as_(col_field.name),
        )
        return StepContext(query, columns)

    def top(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "TopStep",
    ) -> StepContext:
        if step.groups:
            if self.SUPPORT_ROW_NUMBER:
                sub_query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(*columns)

                the_table = Table(prev_step_name)
                rank_on_field: Field = the_table[step.rank_on]
                groups_fields: list[Field] = [the_table[group] for group in step.groups]
                sub_query = sub_query.select(
                    RowNumber()
                    .over(*groups_fields)
                    .orderby(rank_on_field, order=Order.desc if step.sort == "desc" else Order.asc)
                    .as_("row_number")
                )
                query: "QueryBuilder" = (
                    self.QUERY_CLS.from_(sub_query)
                    .select(*columns)
                    .where(Field("row_number") <= step.limit)
                    # The order of returned results is not necessarily consistent. This ensures we
                    # always get the results in the same order
                    .orderby(*(Field(f) for f in step.groups + ["row_number"]), order=Order.asc)
                )
                return StepContext(query, columns)

            else:
                raise NotImplementedError(f"[{self.DIALECT}] top is not implemented with groups")

        query = (
            self.QUERY_CLS.from_(prev_step_name)
            .select(*columns)
            .orderby(step.rank_on, order=Order.desc if step.sort == "desc" else Order.asc)
            .limit(step.limit)
        )
        return StepContext(query, columns)

    def trim(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "TrimStep",
    ) -> StepContext:
        col_fields: list[Field] = [Table(prev_step_name)[col] for col in step.columns]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c not in step.columns),
            *(functions.Trim(col_field).as_(col_field.name) for col_field in col_fields),
        )
        return StepContext(query, columns)

    def uniquegroups(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "UniqueGroupsStep",
    ) -> StepContext:
        from weaverbird.pipeline.steps import AggregateStep

        return self.aggregate(
            step=AggregateStep(on=step.on, aggregations=[], keepOriginalGranularity=False),
            builder=builder,
            prev_step_name=prev_step_name,
            columns=columns,
        )

    @classmethod
    def _build_unpivot_col(
        cls, *, step: 'UnpivotStep', quote_char: str | None, secondary_quote_char: str
    ) -> str:
        value_col = format_quotes(step.value_column_name, quote_char)
        unpivot_col = format_quotes(step.unpivot_column_name, quote_char)
        in_cols = ', '.join(format_quotes(col, quote_char) for col in step.unpivot)
        if cls.SUPPORT_UNPIVOT:
            return f'UNPIVOT({value_col} FOR {unpivot_col} IN ({in_cols}))'
        in_single_quote_cols = ', '.join(
            format_quotes(col, secondary_quote_char) for col in step.unpivot
        )
        return f' t1 CROSS JOIN UNNEST(ARRAY[{in_single_quote_cols}], ARRAY[{in_cols}]) t2 ({unpivot_col}, {value_col})'

    def unpivot(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: 'UnpivotStep',
    ) -> StepContext:
        unpivot = self._build_unpivot_col(
            step=step,
            quote_char=builder.QUOTE_CHAR,
            secondary_quote_char=builder.SECONDARY_QUOTE_CHAR,
        )
        cols = step.keep + [step.unpivot_column_name] + [step.value_column_name]
        query = LiteralValue(f'{self.QUERY_CLS.from_(prev_step_name).select(*cols)!s} {unpivot}')
        return StepContext(query, cols)

    def uppercase(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "UppercaseStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *(c for c in columns if c != step.column),
            functions.Upper(col_field).as_(step.column),
        )
        return StepContext(query, columns)


class CountDistinct(functions.Count):  # type: ignore[misc]
    def __init__(self, param: str | Field, alias: str | None = None) -> None:
        super().__init__(param, alias)
        self._distinct = True


class DateFormat(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("DATE_FORMAT", term, date_format, alias=alias)


class RowNumber(AnalyticFunction):  # type: ignore[misc]
    def __init__(self, **kwargs: Any) -> None:
        super().__init__("ROW_NUMBER", **kwargs)


class StrToDate(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("STR_TO_DATE", term, date_format, alias=alias)


class ParseDate(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("PARSE_DATE", term, date_format, alias=alias)


class ToTimestampNTZ(functions.Function):  # type: ignore[misc]
    def __init__(self, term: str | Field, date_format: str, alias: str | None = None) -> None:
        super().__init__("TO_TIMESTAMP_NTZ", term, date_format, alias=alias)


class RegexpMatching(Comparator):  # type: ignore[misc]
    similar_to = " SIMILAR TO "
    not_similar_to = " NOT SIMILAR TO "
    contains = " CONTAINS "
    not_contains = " NOT CONTAINS "
    regexp = " REGEXP "


def _compliant_regex(pattern: str, dialect: SQLDialect) -> str:
    """
    Like LIKE, the SIMILAR TO operator succeeds only if its pattern matches the entire string;
    this is unlike common regular expression behavior wherethe pattern
    can match any part of the string
    (see https://www.postgresql.org/docs/current/functions-matching.html#FUNCTIONS-SIMILARTO-REGEXP)

    For some special cases like googlebigquery or athena, we don't need to add
    those %
    """

    if dialect in (SQLDialect.ATHENA, SQLDialect.GOOGLEBIGQUERY):
        return f"{pattern}"
    elif dialect in (SQLDialect.SNOWFLAKE, SQLDialect.MYSQL):
        return f".*{pattern}.*"

    return f"%{pattern}%"



from pypika.dialects import Query
from pypika.terms import CustomFunction, Field, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DATE_INFO,
    DataTypeMapping,
    SQLTranslator,
)


class AthenaTranslator(SQLTranslator):
    DIALECT = SQLDialect.ATHENA
    QUERY_CLS = Query
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = False
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="FLOAT",
        integer="INTEGER",
        text="VARCHAR",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    REGEXP_OP = RegexOp.REGEXP_LIKE
    TO_DATE_OP = ToDateOp.TIMESTAMP

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        add_date_func = CustomFunction('DATE_ADD', ['interval', 'increment', 'datecol'])
        return add_date_func(add_date_unit, add_date_value, date_column)


SQLTranslator.register(AthenaTranslator)



import operator
from ast import literal_eval

from pypika import Table
from pypika.functions import NullIf
from pypika.terms import Term

from weaverbird.pipeline.formula_ast.eval import FormulaParser
from weaverbird.pipeline.formula_ast.types import ColumnName, Expression, Operation, Operator

_OP_MAP = {
    Operator.ADD: operator.add,
    Operator.SUB: operator.sub,
    Operator.MUL: operator.mul,
    Operator.DIV: operator.truediv,
    Operator.MOD: operator.mod,
}


def _eval_operation(op: Operation, table: Table) -> Term:
    right = (
        NullIf(_eval_expression(op.right, table), 0)
        if op.operator in (Operator.DIV, Operator.MOD)
        else _eval_expression(op.right, table)
    )
    return _OP_MAP[op.operator](_eval_expression(op.left, table), right)


def _eval_expression(expr: Expression, table: Table) -> Term:
    if isinstance(expr, Operation):
        return _eval_operation(expr, table)
    elif isinstance(expr, ColumnName):
        return table[expr.name]
    elif isinstance(expr, str):
        # unquoting
        expr = literal_eval(expr)
    return Term.wrap_constant(expr)


def formula_to_term(formula: str, table: Table) -> Term:
    return _eval_expression(FormulaParser(formula).parse(), table)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import DeleteStep


def execute_delete(
    step: DeleteStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    return df.drop(step.columns, axis=1, errors='ignore')



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ReplaceStep


def translate_replace(step: ReplaceStep) -> List[MongoStep]:
    branches: List[MongoStep] = [
        {
            'case': {'$eq': [f'${step.search_column}', old_value]},
            'then': new_value,
        }
        for (old_value, new_value) in step.to_replace
    ]
    return [
        {
            '$addFields': {
                f'{step.search_column}': {
                    '$switch': {'branches': branches, 'default': f'${step.search_column}'},
                }
            }
        }
    ]



from typing import List, Literal, Union

from pydantic import BaseModel


class Reference(BaseModel):
    type: Literal['ref'] = 'ref'
    uid: str

    # Adding a prefix so is does not have the same hash as a domain whose name would be the same as
    # the uid
    def __hash__(self):
        return hash(f'__ref__{self.uid}')


PipelineOrDomainName = Union[List[dict], str]  # can be either a domain name or a complete pipeline
PipelineOrDomainNameOrReference = Union[PipelineOrDomainName, Reference]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName


class AbsoluteValueStep(BaseStep):
    name: Literal['absolutevalue'] = 'absolutevalue'
    column: ColumnName
    new_column: ColumnName


class AbsoluteValueStepWithVariable(AbsoluteValueStep, StepWithVariablesMixin):
    ...



from ast import literal_eval
from enum import Enum
from typing import Union

from pydantic import BaseModel


class Operator(str, Enum):
    ADD = '+'
    SUB = '-'
    MUL = '*'
    DIV = '/'
    MOD = '%'


Constant = Union[int, bool, float, str]


class ColumnName(BaseModel):
    name: str
    alias: str


Expression = Union['Operation', ColumnName, Constant]


# Dataclasses do not supported recursive types for now
class Operation(BaseModel, smart_union=True):
    left: Expression
    right: Expression
    operator: Operator


Operation.update_forward_refs()


def format_expr(
    expr: Expression,
    *,
    parenthesize_operations: bool = False,
    str_quote_seq: str = "'",
    column_start_seq: str = '[',
    column_end_seq: str = ']',
    bools_as_py: bool = False,
) -> str | int | float:
    if isinstance(expr, Operation):
        left = format_expr(
            expr.left,
            parenthesize_operations=True,
            str_quote_seq=str_quote_seq,
            column_start_seq=column_start_seq,
            column_end_seq=column_end_seq,
        )
        right = format_expr(
            expr.right,
            parenthesize_operations=True,
            str_quote_seq=str_quote_seq,
            column_start_seq=column_start_seq,
            column_end_seq=column_end_seq,
        )
        str_ = f'{left} {expr.operator} {right}'
        return f'({str_})' if parenthesize_operations else str_
    elif isinstance(expr, ColumnName):
        return f'{column_start_seq}{expr.name}{column_end_seq}'
    elif isinstance(expr, str):
        # Unquoting
        return f"{str_quote_seq}{literal_eval(expr)}{str_quote_seq}"
    # bool is a subtype of int, but we don't want to have 1 and 0 replaced with 'true' or 'false'
    elif isinstance(expr, (int, float)) and not isinstance(expr, bool):
        return expr
    elif isinstance(expr, bool):
        return expr if bools_as_py else ('true' if expr else 'false')



from typing import List, Literal, Union

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import ColumnName, TemplatedVariable

DatesGranularity = Union[Literal['day'], Literal['week'], Literal['month'], Literal['year']]


class AddMissingDatesStep(BaseStep):
    name: Literal['addmissingdates'] = 'addmissingdates'
    dates_column: ColumnName = Field(alias='datesColumn')
    dates_granularity: DatesGranularity = Field(alias='datesGranularity')
    groups: List[ColumnName] = []


class AddMissingDatesStepWithVariables(AddMissingDatesStep, StepWithVariablesMixin):
    groups: Union[List[TemplatedVariable], TemplatedVariable]
    dates_column: TemplatedVariable = Field(alias='datesColumn')



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.combination import (
    resolve_sql_pipeline_for_combination,
)
from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
    build_union_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import AppendStep


def translate_append(
    step: AppendStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'APPEND_STEP_{index}'
    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )
    queries_to_append = {}
    query_to_union_metadata = {}

    transformed_query = f'{query.transformed_query}'

    for index, pipeline in enumerate(step.pipelines):
        query_string = resolve_sql_pipeline_for_combination(
            pipeline_or_domain=pipeline,
            sql_query_retriever=sql_query_retriever,
            sql_translate_pipeline=sql_translate_pipeline,
            sql_query_describer=sql_query_describer,
            sql_query_executor=sql_query_executor,
            subcall_from_other_pipeline_count=index,
        )
        unioned_query_name = f'APPEND_STEP_UNION_{index}'
        queries_to_append[unioned_query_name] = query_string
        query_to_union_metadata[unioned_query_name] = sql_query_describer(
            domain=None, query_string=query_string
        )
        log.debug(
            '------------------------------------------------------------'
            f'SQLquery: {transformed_query}'
            '############################################################'
        )
        query.metadata_manager.create_table(unioned_query_name)
        query.metadata_manager.add_table_columns_from_dict(
            unioned_query_name, query_to_union_metadata[unioned_query_name]
        )
        transformed_query += f', {unioned_query_name} AS ({query_string})'

    query.metadata_manager.append_queries_metadata(unioned_tables=queries_to_append.keys())
    transformed_query += f', {query_name} AS\
 ({build_union_query(query.metadata_manager, query.query_name, queries_to_append.keys())})'
    query.metadata_manager.rename_union_columns()

    return SQLQuery(
        query_name=query_name,
        transformed_query=transformed_query,
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import DomainStep


def translate_domain(step: DomainStep) -> List[MongoStep]:
    return [{'$match': {'domain': step.domain}}]



import logging
from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.backends.mongo_translator.utils import column_to_user_variable
from weaverbird.pipeline import Pipeline, steps
from weaverbird.pipeline.steps import DomainStep, JoinStep

logger = logging.getLogger(__name__)


def translate_join(step: JoinStep) -> List[MongoStep]:
    mongo_pipeline: List[dict] = []
    right = step.right_pipeline
    right_without_domain = Pipeline(steps=[])

    if isinstance(right, str):
        right_domain = DomainStep(name='domain', domain=right)
    else:
        try:
            assert isinstance(right, list)
        except AssertionError:  # in this case right is a Reference
            raise Exception('References must be resolved before translating the pipeline')
        right_domain = DomainStep(**right[0])
        right_without_domain.steps = [
            getattr(steps, f"{s['name'].capitalize()}Step")(**s) for s in right[1:]
        ]

    mongo_let: dict[str, str] = {}
    mongo_expr_and: list[dict[str, list[str]]] = []

    for left_on, right_on in step.on:
        mongo_let[column_to_user_variable(left_on)] = f'${left_on}'
        mongo_expr_and.append({'$eq': [f'${right_on}', f'$${column_to_user_variable(left_on)}']})

    from weaverbird.backends.mongo_translator.mongo_pipeline_translator import translate_pipeline

    mongo_pipeline.append(
        {
            '$lookup': {
                'from': right_domain.domain,
                'let': mongo_let,
                'pipeline': translate_pipeline(right_without_domain)
                + [{'$match': {'$expr': {'$and': mongo_expr_and}}}],
                'as': '_vqbJoinKey',
            }
        }
    )
    if step.type == 'inner':
        mongo_pipeline.append({'$unwind': '$_vqbJoinKey'})
    elif step.type == 'left':
        mongo_pipeline.append(
            {'$unwind': {'path': '$_vqbJoinKey', 'preserveNullAndEmptyArrays': True}}
        )
    else:
        mongo_pipeline.append({'$match': {'_vqbJoinKey': {'$eq': []}}})
        mongo_pipeline.append(
            {'$unwind': {'path': '$_vqbJoinKey', 'preserveNullAndEmptyArrays': True}}
        )

    mongo_pipeline.append(
        {
            '$replaceRoot': {'newRoot': {'$mergeObjects': ['$_vqbJoinKey', '$$ROOT']}},
        }
    )
    mongo_pipeline.append(
        {
            '$project': {'_vqbJoinKey': 0},
        },
    )
    return mongo_pipeline



from typing import TYPE_CHECKING, Any, TypeVar

from pypika import Field, Query, Table, functions
from pypika.queries import QueryBuilder
from pypika.terms import LiteralValue, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import (
    DATE_INFO,
    DataTypeMapping,
    SQLTranslator,
    StepContext,
)

Self = TypeVar("Self", bound="GoogleBigQueryTranslator")

if TYPE_CHECKING:
    from weaverbird.pipeline.steps import SplitStep


class GoogleBigQueryQuery(Query):  # type: ignore[misc]
    @classmethod
    def _builder(cls, **kwargs: Any) -> "GoogleBigQueryQueryBuilder":
        return GoogleBigQueryQueryBuilder(**kwargs)


class GoogleBigQueryQueryBuilder(QueryBuilder):  # type: ignore[misc]
    QUOTE_CHAR = "`"
    SECONDARY_QUOTE_CHAR = "'"
    ALIAS_QUOTE_CHAR = None
    QUERY_ALIAS_QUOTE_CHAR = None
    QUERY_CLS = GoogleBigQueryQuery

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(dialect=SQLDialect.GOOGLEBIGQUERY, **kwargs)


class GoogleBigQueryTranslator(SQLTranslator):
    DIALECT = SQLDialect.GOOGLEBIGQUERY
    QUERY_CLS = GoogleBigQueryQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="STRING",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = False
    SUPPORT_UNPIVOT = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP_CONTAINS
    TO_DATE_OP = ToDateOp.TIMESTAMP

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        return LiteralValue(
            f"DATE_ADD({date_column.name}, INTERVAL {add_date_value} {add_date_unit})"
        )

    def split(
        self: Self,
        *,
        builder: 'QueryBuilder',
        prev_step_name: str,
        columns: list[str],
        step: "SplitStep",
    ) -> StepContext:
        col_field: Field = Table(prev_step_name)[step.column]
        splitted_cols = [
            LiteralValue(f'SPLIT(`{col_field.name}`, "{step.delimiter}")[SAFE_OFFSET({i})]').as_(
                f"{step.column}_{i + 1}"
            )
            for i in range(step.number_cols_to_keep)
        ]
        query: "QueryBuilder" = self.QUERY_CLS.from_(prev_step_name).select(
            *columns, *splitted_cols
        )
        return StepContext(query, columns + splitted_cols)


SQLTranslator.register(GoogleBigQueryTranslator)


class ParseDatetime(functions.Function):  # type: ignore[misc]
    def __init__(self, format: str, term: str | Field, alias: str | None = None) -> None:
        super().__init__("parse_datetime", format, term, alias=alias)



from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import PercentageStep


def execute_percentage(
    step: PercentageStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    new_column_name = step.new_column_name or f'{step.column}_PCT'

    if len(step.group) > 0:
        sums = df.groupby(step.group, dropna=False)[step.column].transform('sum')
    else:
        sums = df[step.column].sum()
    return df.assign(**{new_column_name: df[step.column] / sums})



class WeaverbirdError(Exception):
    """Base class for weaverbird exceptions"""


class DuplicateError(WeaverbirdError):
    """Raised when an operation encountered duplicates where it was forbidden"""


class DuplicateColumnError(Exception):
    """raised when duplicate column names in input"""


class UnresolvedReferenceError(Exception):
    """Raised when a join step as a reference instead of a pipeline in input"""


class MissingColumnNameError(WeaverbirdError):
    """Raised when defining a substring but did not supply a new column name"""


class MissingTableNameError(WeaverbirdError):
    """Raised when defining a custom SQL query but didn't set a proper temporary table name"""



from typing import Any, List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import AggregateStep


def translate_aggregate(step: AggregateStep) -> List[MongoStep]:
    idblock = {col: f'${col}' for col in step.on}
    group: dict[str, dict] = {}
    project: dict[str, Any] = {}
    add_fields = {}
    group['_id'] = idblock

    for agg_f_step in step.aggregations:
        cols = agg_f_step.columns
        new_cols = agg_f_step.new_columns

        # There is no `$count` operator in Mongo, we have to `$sum` 1s to get
        # an equivalent result
        if agg_f_step.agg_function == 'count':
            for i in range(len(cols)):
                group[new_cols[i]] = {'$sum': 1}
        elif agg_f_step.agg_function == 'count distinct':
            for i in range(len(cols)):
                group[new_cols[i]] = {'$addToSet': f'${cols[i]}'}
                add_fields[new_cols[i]] = {'$size': f'${new_cols[i]}'}
        else:
            for i in range(len(cols)):
                group[new_cols[i]] = {f'${agg_f_step.agg_function}': f'${cols[i]}'}

    add_fields_to_add_to_pipeline = [{'$addFields': add_fields}] if len(add_fields) else []
    if step.keep_original_granularity:
        group['_vqbDocsArray'] = {'$push': '$$ROOT'}
        return [
            {'$group': group},
            *add_fields_to_add_to_pipeline,
            {'$unwind': '$_vqbDocsArray'},
            {'$replaceRoot': {'newRoot': {'$mergeObjects': ['$_vqbDocsArray', '$$ROOT']}}},
            {'$project': {'_vqbDocsArray': 0}},
            {'$sort': {'_id': 1}},
        ]
    else:
        for group_key in group.keys():
            if group_key == '_id':
                for id_key in group[group_key]:
                    project[id_key] = f'$_id.{id_key}'
            else:
                project[group_key] = 1
        return [
            {'$group': group},
            *add_fields_to_add_to_pipeline,
            {'$project': project},
            {'$sort': {'_id': 1}},
        ]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import UniqueGroupsStep


def translate_uniquegroups(step: UniqueGroupsStep) -> List[MongoStep]:
    id_block = {col: f"${col}" for col in step.on}
    project_block = {col: f"$_id.{col}" for col in step.on}
    return [{'$group': {'_id': id_block}}, {'$project': project_block}]



from typing import Literal

from pydantic import validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin


class CustomSqlStep(BaseStep):
    name: Literal['customsql'] = 'customsql'
    query: str

    @staticmethod
    def _strip_query(query: str) -> str:
        return query.strip().strip(';')

    @validator('query')
    def _validate_query(cls, query: str) -> str:
        assert ';' not in (
            stripped := cls._strip_query(query)
        ), "Custom SQL queries must not contain semicolumns"
        return stripped


class CustomSqlStepWithVariables(CustomSqlStep, StepWithVariablesMixin):
    ...



import ast
import tokenize
from io import BytesIO
from typing import Generator, Iterator

from . import types


class ASTError(Exception):
    """Base exception for all AST errors"""


class UnclosedColumnName(ASTError):
    """Column name was not closed"""


class EmptyColumnName(ASTError):
    """Column name was not closed"""


class UnexpectedToken(ASTError):
    """Unexpected token"""


class InvalidFormula(ASTError):
    """Invalid formula"""


class UnsupportedOperator(ASTError):
    """Unsupported operator"""


class UnsupportedExpression(ASTError):
    """Unsupported expression"""


class UnsupportedConstant(ASTError):
    """Unsupported constant"""


class FormulaParser:
    def __init__(self, formula: str) -> None:
        self._formula = formula
        self._columns: dict[str, str] = {}

    @staticmethod
    def sanitize_string(s: str) -> str:
        """Unquotes strings, escapes single quotes and wraps them in single quotes"""
        # Removing outer quotes
        unquoted = ast.literal_eval(s)
        # Escaping single quotes in the string
        with_escaped_single_quotes = unquoted.replace('\'', r'\'')
        # Wrapping the entire thing in single quotes
        return f"'{with_escaped_single_quotes}'"

    def _iterate_tokens(self, tokens: Iterator[tokenize.TokenInfo]) -> Generator[str, None, None]:
        """This function iterates over tokens, sanitizes and yields them."""

        def next_token():
            try:
                return next(tokens)
            except StopIteration:
                return None

        def parse_col_name(prev_token: tokenize.TokenInfo) -> str:
            chunks = []
            # Skipping the current token, which is [
            while tok := next_token():
                # We have an identifier, append it
                if tok.type == tokenize.NAME:
                    chunks.append(tok.string)
                # We reached the end of the column name. Store an alias for it and yield the alias
                elif tok.type == tokenize.OP and tok.string == ']':
                    if len(chunks) < 1:
                        raise EmptyColumnName(f'Got an empty column name at {prev_token.start[1]}')
                    col_name = f'__VQB_COL__{len(self._columns)}'
                    self._columns[col_name] = ' '.join(chunks)
                    return col_name
                else:
                    raise UnexpectedToken(f"Unexpected token {tok} in column name")
                prev_token = tok
            raise UnclosedColumnName(
                f'Expected column to be closed near {prev_token.string} at {prev_token.start[1]}'
            )

        while token := next_token():
            if token.type in (
                tokenize.ENCODING,
                tokenize.INDENT,
                tokenize.NEWLINE,
                tokenize.ENDMARKER,
            ):
                # Those are whitespace tokens we don't care about, so not yielding them
                continue
            elif token.type == tokenize.STRING:
                # In case we have a string literal, we sanitize it
                yield self.sanitize_string(token.string)
            elif token.type == tokenize.OP and token.string == '[':
                # Square brackets are delimiters for column names, so here we read everyhting until
                # the closing bracket
                yield parse_col_name(token)
            else:
                yield token.string

    def sanitize_formula(self) -> str:
        """Removes [] around column names and aliases them.

        This is needed because stuff parsed with the STL's ast module needs to be valid python code
        """
        self._columns = {}
        # Stripping because strings starting with whitespace raise UnexpectedIndent when parsed by
        # the ast module
        return ' '.join(
            self._iterate_tokens(tokenize.tokenize(BytesIO(self._formula.encode()).readline))
        ).strip()

    @staticmethod
    def _operator_from_ast_op(op: ast.operator) -> types.Operator:
        """Returns one of our known operators from an operator of the ast module.

        Raise an exception in case of an unsupported operator.
        """
        match op:
            case ast.Add():
                return types.Operator.ADD
            case ast.Sub():
                return types.Operator.SUB
            case ast.Mult():
                return types.Operator.MUL
            case ast.Div():
                return types.Operator.DIV
            case ast.Mod():
                return types.Operator.MOD
            case _:
                raise UnsupportedOperator(f'Unsupported operator: {op}')

    def _build_name(self, name: str) -> types.ColumnName | bool:
        """Builds a ColumnName from a raw name.

        In case the name is "true" or "false", returns the equivalent boolean
        (users do not expect to have to capitalize booleans).
        """
        if name == 'true':
            return True
        if name == 'false':
            return False
        return (
            types.ColumnName(name=self._columns[name], alias=name)
            if name in self._columns
            else types.ColumnName(name=name, alias=name)
        )

    def _parse_expr(self, expr: ast.expr) -> types.Expression:
        """Takes an expression of the ast module and builds an AST in our own format from it."""
        match expr:
            # +LITERAL: +1, +13.37
            case ast.UnaryOp(op=ast.UAdd(), operand=ast.Constant(value=value)):
                return value
            # -LITERAL: -1, -13.37
            case ast.UnaryOp(op=ast.USub(), operand=ast.Constant(value=value)):
                return -value
            # +colname: +mycol, +[my col]
            case ast.UnaryOp(op=ast.UAdd(), operand=ast.Name(id=name)):
                return self._build_name(name)
            # -colname: -mycol, -[my col]
            case ast.UnaryOp(op=ast.USub(), operand=ast.Name(id=name)):
                # Cheating a bit here, assuming the column is numeric
                return types.Operation(
                    left=-1, operator=types.Operator.MUL, right=self._build_name(name)
                )
            # Recursing down into both branches of the operation
            case ast.BinOp(left=left, right=right, op=op):
                operator = self._operator_from_ast_op(op)
                return types.Operation(
                    left=self._parse_expr(left), right=self._parse_expr(right), operator=operator
                )
            # Constant: number, string literal or boolean
            case ast.Constant(value=value):
                # bool is a subtype of int
                if isinstance(value, (int, float)):
                    return value
                elif isinstance(value, str):
                    return f"'{value}'"
                else:
                    raise UnsupportedConstant(
                        f"Unsupported constant '{expr}' of type {type(value)}"
                    )
            # Column name
            case ast.Name(id=name):
                return self._build_name(name)
            case _:
                raise UnsupportedExpression(f"Unsupported expression: {ast.dump(expr)}")

    def parse(self) -> types.Expression:
        parsed = ast.parse(self.sanitize_formula())
        match parsed:
            case ast.Module(body=[ast.Expr(value=expr)]):
                return self._parse_expr(expr)
            case _:
                raise InvalidFormula("Invalid formula: Expected an expression at root")



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.addmissingdates import AddMissingDatesStep, DatesGranularity


def _add_missing_dates_year(step: AddMissingDatesStep) -> List[MongoStep]:
    groups = step.groups or []
    # add missing dates by looping over the unique dates array and adding a given date
    # if it's not already present in the original dataset
    add_missing_years_as_dates: MongoStep = {
        '$map': {
            # loop over a sorted array of all years between min and max year
            'input': {'$range': ['$_vqbMinYear', {'$add': ['$_vqbMaxYear', 1]}]},
            # use a variable "currentYear" as cursor
            'as': 'currentYear',
            # and apply the following expression to every "currentYear"
            'in': {
                '$let': {
                    # use a variable yearIndex that represents the index of the "currentYear"
                    # cursor in the original array
                    'vars': {
                        'yearIndex': {'$indexOfArray': ['$_vqbArray._vqbYear', '$$currentYear']},
                    },
                    'in': {
                        '$cond': [
                            # if "currentYear" is found in the original array
                            {'$ne': ['$$yearIndex', -1]},
                            # just get the original document in the original array
                            {'$arrayElemAt': ['$_vqbArray', '$$yearIndex']},
                            # else add a new document with the missing date (we convert the year back to a date object)
                            # and the group fields (every other field will be undefined)
                            {
                                **{col: f'$_id.{col}' for col in groups},
                                step.dates_column: {'$dateFromParts': {'year': '$$currentYear'}},
                            },
                        ],
                    },
                }
            },
        }
    }

    return [
        # Extract the year of the date in a new column
        {'$addFields': {'_vqbYear': {'$year': f'${step.dates_column}'}}},
        # Group by logic to create an array with all original documents, and get the
        # min and max date of all documents confounded
        {
            '$group': {
                '_id': {col: f'${col}' for col in groups} if groups else None,
                '_vqbArray': {'$push': '$$ROOT'},
                '_vqbMinYear': {'$min': '$_vqbYear'},
                '_vqbMaxYear': {'$max': '$_vqbYear'},
            },
        },
        # That at this stage that all the magic happens (cf. supra)
        {'$addFields': {'_vqbAllDates': add_missing_years_as_dates}},
    ]


def _generate_date_from_parts(mongo_date: str, granularity: DatesGranularity) -> MongoStep:
    date_from_parts = {'year': {'$year': mongo_date}}
    if granularity in ('day', 'month'):
        date_from_parts['month'] = {'$month': mongo_date}
    if granularity == 'day':
        date_from_parts['day'] = {'$dayOfMonth': mongo_date}

    return date_from_parts


def _add_missing_dates_day_or_month(step: AddMissingDatesStep) -> List[MongoStep]:
    groups = step.groups or []
    # Create a sorted array of all dates (in days) ranging from min to
    # max dates found in the whole dataset. At this stage for the
    # month granularity, we will get duplicate dates because we need
    # to first generate all days one by one to increment the calendar
    # safely, before converting every day of a given  month to the 1st
    # of this month (via the $dateFromPart expression). We take care
    # of making month dates unique in a '$reduce' stage explained
    # later
    all_days_range = {
        '$map': {
            # create a array of integers ranging from 0 to the number of days separating the min and max date
            'input': {'$range': [0, {'$add': ['$_vqbMinMaxDiffInDays', 1]}]},
            # use a cursor 'currentDurationInDays' that will loop over the array
            'as': 'currentDurationInDays',
            # apply the following operations to every element of the array via the cursor 'currentDurationInDays'
            'in': {
                '$let': {
                    # create a variable 'currentDay' that will correspond to the day resulting from the addition of the
                    # duration 'currentDurationInDays' (converted back to milliseconds) to the min date.
                    'vars': {
                        'currentDay': {
                            '$add': [
                                '$_vqbMinDay',
                                {'$multiply': ['$$currentDurationInDays', 60 * 60 * 24 * 1000]},
                            ],
                        },
                    },
                    # use the variable in the following expression, in which we recreate a date which granularity will
                    # depend on the user-specified granularity
                    'in': {
                        '$dateFromParts': _generate_date_from_parts(
                            '$$currentDay', step.dates_granularity
                        ),
                    },
                },
            },
        },
    }

    # For the month granularity only, use a $reduce stage to reduce the allDaysRange
    # array into an array of unique days.
    unique_days_for_month_granularity = {
        '$reduce': {
            'input': all_days_range,
            'initialValue': [],
            'in': {
                '$cond': [
                    # if the date is not found in the reduced array
                    {'$eq': [{'$indexOfArray': ['$$value', '$$this']}, -1]},
                    # then add the date to it
                    {'$concatArrays': ['$$value', ['$$this']]},
                    # else add nothing to it
                    {'$concatArrays': ['$$value', []]},
                ],
            },
        },
    }

    # add missing dates by looping over the unique dates array and adding a given date
    # if it's not already present in the original dataset
    add_missing_dates = {
        '$map': {
            # loop over unique dates array
            'input': all_days_range
            if step.dates_granularity == 'day'
            else unique_days_for_month_granularity,
            # use a variable "date" as cursor
            'as': 'date',
            # and apply the following expression to every "date"
            'in': {
                '$let': {
                    # use a variable dateIndex that represents the index
                    # of the "date" cursor in the original array of documents
                    'vars': {'dateIndex': {'$indexOfArray': ['$_vqbArray._vqbDay', '$$date']}},
                    'in': {
                        '$cond': [
                            # if "date" is found in the original array
                            {'$ne': ['$$dateIndex', -1]},
                            # just get the original document in the original array
                            {'$arrayElemAt': ['$_vqbArray', '$$dateIndex']},
                            # else add a new document with the missing
                            # date and the group fieldds (every other
                            # field will be undefined)
                            {
                                **{col: f'$_id.{col}' for col in groups},
                                step.dates_column: '$$date',
                            },
                        ],
                    },
                },
            },
        },
    }

    return [
        {
            '$addFields': {
                '_vqbDay': {
                    '$dateFromParts': _generate_date_from_parts(
                        f'${step.dates_column}', step.dates_granularity
                    ),
                },
            },
        },
        # Group by logic to create an array with all original documents, and get the
        # min and max date of all documents confounded
        {
            '$group': {
                '_id': {col: f'${col}' for col in groups} if step.groups else None,
                # '_id': groups or None,
                '_vqbArray': {'$push': '$$ROOT'},
                '_vqbMinDay': {'$min': '$_vqbDay'},
                '_vqbMaxDay': {'$max': '$_vqbDay'},
            },
        },
        # Compute the time difference between the min and max date in days (the
        # subtraction between two dates in mongo gives a duration in milliseconds)
        {
            '$addFields': {
                '_vqbMinMaxDiffInDays': {
                    '$divide': [{'$subtract': ['$_vqbMaxDay', '$_vqbMinDay']}, 60 * 60 * 24 * 1000],
                },
            },
        },
        # That at this stage that all the magic happens (cf. supra)
        {'$addFields': {'_vqbAllDates': add_missing_dates}},
    ]


def translate_addmissingdates(step: AddMissingDatesStep) -> List[MongoStep]:
    return (
        _add_missing_dates_year(step)
        if step.dates_granularity == 'year'
        else _add_missing_dates_day_or_month(step)
    ) + [
        # Get back to 1 row per document
        {'$unwind': '$_vqbAllDates'},
        # Change the root to get back to the document granularity
        {'$replaceRoot': {'newRoot': '$_vqbAllDates'}},
        # Remove remaining temporary column
        {'$project': {'_vqbYear' if step.dates_granularity == 'year' else '_vqbDay': 0}},
    ]



from typing import Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class LowercaseStep(BaseStep):
    name: Literal['lowercase'] = 'lowercase'
    column: ColumnName



from typing import Any, Dict, List

from pandas import DataFrame

from weaverbird.backends.pandas_executor.geo_utils import df_to_geodf
from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.aggregate import Aggregation
from weaverbird.pipeline.steps.dissolve import DissolveStep

_AGG_FUNC_ALIASES = {
    'avg': 'mean',
    'count distinct': 'nunique',
    'count distinct including empty': len,
}


def _translate_agg_func(aggregations: List[Aggregation]) -> Dict[str, Any]:
    return {
        agg.columns[0]: _AGG_FUNC_ALIASES[agg.agg_function]
        if agg.agg_function in _AGG_FUNC_ALIASES
        else agg.agg_function
        for agg in aggregations
    }


def execute_dissolve(
    step: DissolveStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    if len(step.aggregations) > 0:
        aggfunc: dict[str, Any] | str = _translate_agg_func(step.aggregations)
        geo_df = df_to_geodf(df)
    else:
        aggfunc = 'first'  # placeholder, can't be None
        columns_to_keep = set(step.groups + ['geometry'])
        geo_df = df_to_geodf(df).drop(list(set(df.columns) - columns_to_keep), axis=1)

    return DataFrame(
        geo_df.dissolve(
            by=step.groups,
            aggfunc=aggfunc,
            as_index=False,
            dropna=not step.include_nulls,
        )
    )



from typing import List, Literal

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin

from .utils.combination import PipelineOrDomainNameOrReference


class AppendStep(BaseStep):
    name: Literal['append'] = 'append'
    pipelines: List[PipelineOrDomainNameOrReference]


class AppendStepWithVariable(AppendStep, StepWithVariablesMixin):
    ...



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ArgmaxStep


def translate_argmax(step: ArgmaxStep) -> List[MongoStep]:
    return [
        {
            '$group': {
                '_id': {group: f'${group}' for group in step.groups},
                '_vqbAppArray': {'$push': '$$ROOT'},
                '_vqbAppValueToCompare': {'$max': f'${step.column}'},
            }
        },
        {'$unwind': '$_vqbAppArray'},
        {'$replaceRoot': {'newRoot': {'$mergeObjects': ['$_vqbAppArray', '$$ROOT']}}},
        {'$project': {'_vqbAppArray': 0}},
        {
            '$redact': {
                '$cond': [
                    {'$eq': [f'${step.column}', '$_vqbAppValueToCompare']},
                    '$$KEEP',
                    '$$PRUNE',
                ]
            }
        },
        {'$project': {'_vqbAppValueToCompare': 0}},
    ]



from typing import Any, Callable, Dict

from weaverbird.backends.mongo_translator.steps.absolutevalue import translate_absolutevalue
from weaverbird.backends.mongo_translator.steps.addmissingdates import translate_addmissingdates
from weaverbird.backends.mongo_translator.steps.aggregate import translate_aggregate
from weaverbird.backends.mongo_translator.steps.append import translate_append
from weaverbird.backends.mongo_translator.steps.argmax import translate_argmax
from weaverbird.backends.mongo_translator.steps.argmin import translate_argmin
from weaverbird.backends.mongo_translator.steps.comparetext import translate_comparetext
from weaverbird.backends.mongo_translator.steps.concatenate import translate_concatenate
from weaverbird.backends.mongo_translator.steps.convert import translate_convert
from weaverbird.backends.mongo_translator.steps.cumsum import translate_cumsum
from weaverbird.backends.mongo_translator.steps.custom import translate_custom
from weaverbird.backends.mongo_translator.steps.date_extract import translate_date_extract
from weaverbird.backends.mongo_translator.steps.delete import translate_delete
from weaverbird.backends.mongo_translator.steps.domain import translate_domain
from weaverbird.backends.mongo_translator.steps.duplicate import translate_duplicate
from weaverbird.backends.mongo_translator.steps.duration import translate_duration
from weaverbird.backends.mongo_translator.steps.evolution import translate_evolution
from weaverbird.backends.mongo_translator.steps.fillna import translate_fillna
from weaverbird.backends.mongo_translator.steps.filter import translate_filter
from weaverbird.backends.mongo_translator.steps.formula import translate_formula
from weaverbird.backends.mongo_translator.steps.fromdate import translate_fromdate
from weaverbird.backends.mongo_translator.steps.ifthenelse import translate_ifthenelse
from weaverbird.backends.mongo_translator.steps.join import translate_join
from weaverbird.backends.mongo_translator.steps.lowercase import translate_lowercase
from weaverbird.backends.mongo_translator.steps.moving_average import translate_moving_average
from weaverbird.backends.mongo_translator.steps.percentage import translate_percentage
from weaverbird.backends.mongo_translator.steps.pivot import translate_pivot
from weaverbird.backends.mongo_translator.steps.rank import translate_rank
from weaverbird.backends.mongo_translator.steps.rename import translate_rename
from weaverbird.backends.mongo_translator.steps.replace import translate_replace
from weaverbird.backends.mongo_translator.steps.rollup import translate_rollup
from weaverbird.backends.mongo_translator.steps.select import translate_select
from weaverbird.backends.mongo_translator.steps.sort import translate_sort
from weaverbird.backends.mongo_translator.steps.split import translate_split
from weaverbird.backends.mongo_translator.steps.statistics import translate_statistics
from weaverbird.backends.mongo_translator.steps.substring import translate_substring
from weaverbird.backends.mongo_translator.steps.text import translate_text
from weaverbird.backends.mongo_translator.steps.todate import translate_todate
from weaverbird.backends.mongo_translator.steps.top import translate_top
from weaverbird.backends.mongo_translator.steps.totals import translate_totals
from weaverbird.backends.mongo_translator.steps.trim import translate_trim
from weaverbird.backends.mongo_translator.steps.uniquegroups import translate_uniquegroups
from weaverbird.backends.mongo_translator.steps.unpivot import translate_unpivot
from weaverbird.backends.mongo_translator.steps.uppercase import translate_uppercase
from weaverbird.backends.mongo_translator.steps.waterfall import translate_waterfall

# I would like to have a better type for the callable, but I don't know how to do it.
# each of this function take a particular step as input
# so it is not possible to use `Dict[str, Callable[[BaseStep], list]]
mongo_step_translator: Dict[str, Callable[[Any], list]] = {
    'absolutevalue': translate_absolutevalue,
    'addmissingdates': translate_addmissingdates,
    'aggregate': translate_aggregate,
    'append': translate_append,
    'argmax': translate_argmax,
    'argmin': translate_argmin,
    'comparetext': translate_comparetext,
    'concatenate': translate_concatenate,
    'convert': translate_convert,
    'cumsum': translate_cumsum,
    'custom': translate_custom,
    'dateextract': translate_date_extract,
    'delete': translate_delete,
    'domain': translate_domain,
    'duplicate': translate_duplicate,
    'duration': translate_duration,
    'evolution': translate_evolution,
    'fillna': translate_fillna,
    'filter': translate_filter,
    'formula': translate_formula,
    'fromdate': translate_fromdate,
    'ifthenelse': translate_ifthenelse,  # type: ignore
    'join': translate_join,
    'lowercase': translate_lowercase,
    'movingaverage': translate_moving_average,
    'percentage': translate_percentage,
    'pivot': translate_pivot,
    'rename': translate_rename,
    'rank': translate_rank,
    'replace': translate_replace,
    'rollup': translate_rollup,
    'select': translate_select,
    'sort': translate_sort,
    'split': translate_split,
    'statistics': translate_statistics,
    'substring': translate_substring,
    'text': translate_text,
    'todate': translate_todate,
    'totals': translate_totals,
    'top': translate_top,
    'trim': translate_trim,
    'uniquegroups': translate_uniquegroups,
    'unpivot': translate_unpivot,
    'uppercase': translate_uppercase,
    'waterfall': translate_waterfall,
}



from typing import List, Literal, Optional

from pydantic import BaseModel, Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName, PopulatedWithFieldnames

DUMB_GROUPBY_COLUMN_NAME = '__dumb_groupby_column_name__'

Statistics = Literal['count', 'max', 'min', 'average', 'variance', 'standard deviation']


class Quantile(BaseModel):
    label: Optional[str]
    nth: int
    order: int


class StatisticsStep(BaseStep):
    class Config(PopulatedWithFieldnames):
        ...

    name: Literal['statistics'] = 'statistics'
    column: ColumnName
    groupby_columns: List[ColumnName] = Field([], alias='groupbyColumns')
    statistics: List[Statistics]
    # Array of quantiles. Examples:
    # - median is 1rst quantile of order 2
    # - last decile is 9th quantile of order 10
    quantiles: List[Quantile]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import CumSumStep


def translate_cumsum(
    step: CumSumStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'CUMSUM_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    # the partition by sub query
    partition_by_sub_query = 'NULL'
    # if there is a group by query
    if step.groupby is not None and len(step.groupby):
        partition_by_sub_query = ', '.join(step.groupby)

    new_columns = []
    cumsum_part = ""
    for cumsum in step.to_cumsum:
        # if any new column name had been provided
        new_column = cumsum[1] or f"{cumsum[0]}_CUMSUM"

        new_columns.append(new_column)

        cumsum_part += f", SUM({cumsum[0]}) OVER (PARTITION BY {partition_by_sub_query}"
        cumsum_part += (
            f" ORDER BY {step.reference_column} ASC rows UNBOUNDED PRECEDING) {new_column}"
        )

    # we complete fields without the new column name
    # if there is a same naming
    completed_fields = query.metadata_manager.retrieve_query_metadata_columns_as_str(
        columns_filter=new_columns
    )

    # The final query
    final_query = (
        f"SELECT {completed_fields}{cumsum_part}"
        f" FROM {query.query_name} ORDER BY {step.reference_column} ASC"
    )

    # we make sure to add the new columns in the metadata list
    for new_column in new_columns:
        query.metadata_manager.add_query_metadata_column(new_column, "FLOAT")

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"""{query.transformed_query}, {query_name} AS ({final_query})""",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.pivot import PivotStep


def translate_pivot(step: PivotStep) -> List[MongoStep]:
    group_cols = {}
    add_fields_step = {}

    # Prepare groupCols to populate the `_id` field sof Mongo `$group` steps and addFields step
    for col in step.index:
        group_cols[col] = f'$_id.{col}'
        add_fields_step[f'_vqbAppTmpObj.{col}'] = f'$_id.{col}'

    pivot_mongo_agg_stages: List[MongoStep] = [
        # First we perform the aggregation with the _id including the column to pivot
        {
            '$group': {
                '_id': {
                    **{col: f"${col}" for col in step.index},
                    step.column_to_pivot: f'${step.column_to_pivot}',
                },
                step.value_column: {f'${step.agg_function}': f'${step.value_column}'},
            },
        },
        # Then we group with with index columns as _id and we push documents as an array of sets
        # including a column for the column to pivot and a column for the corresponding value
        {
            '$group': {
                '_id': group_cols,
                '_vqbAppArray': {
                    '$addToSet': {
                        step.column_to_pivot: f'$_id.{step.column_to_pivot}',
                        step.value_column: f'${step.value_column}',
                    },
                },
            },
        },
        # Then we project a tmp key to get an object from the array of couples [column_to_pivot, corresponding_value]
        # including a column for the column to pivot and a column for the corresponding value
        {
            '$project': {
                '_vqbAppTmpObj': {
                    '$arrayToObject': {
                        '$zip': {
                            'inputs': [
                                f'$_vqbAppArray.{step.column_to_pivot}',
                                f'$_vqbAppArray.{step.value_column}',
                            ],
                        },
                    },
                },
            },
        },
    ]

    if add_fields_step:
        # Then we include the index columns back in every document created in the previous step
        # (still accessible in the _id object)
        pivot_mongo_agg_stages.append({'$addFields': add_fields_step})

    # Then we replace the root of the documents tree to get our columns ready for
    # our needed table-like, unnested format
    pivot_mongo_agg_stages.append({'$replaceRoot': {'newRoot': '$_vqbAppTmpObj'}})

    return pivot_mongo_agg_stages



import numpy
import pandas as pd
from pandas import DataFrame

from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps.waterfall import (
    _RESULT_COLUMN,
    GROUP_WATERFALL_COLUMN,
    LABEL_WATERFALL_COLUMN,
    TYPE_WATERFALL_COLUMN,
    WaterfallStep,
)


def execute_waterfall(
    step: WaterfallStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    # first milestone
    start_df = df[df[step.milestonesColumn] == step.start].rename(
        columns={step.valueColumn: f'{step.valueColumn}_start'}
    )
    # second milestone
    end_df = df[df[step.milestonesColumn] == step.end].rename(
        columns={step.valueColumn: f'{step.valueColumn}_end'}
    )

    upper = compute_agg_milestone(
        step,
        df,
        start_df.rename(columns={f'{step.valueColumn}_start': step.valueColumn}),
        step.start,
    )
    mid = compute_mid(step, merge(step, start_df, end_df), df)
    downer = compute_agg_milestone(
        step, df, end_df.rename(columns={f'{step.valueColumn}_end': step.valueColumn}), step.end
    )
    return pd.concat([upper, mid, downer])


# start_df is the base dataframe filtered to contains only result at the start of the waterfall
# end_df is the base dataframe filtered to contains only result at the start of the waterfall
# this methods compute the difference for the value value of every label between the end and the start
def merge(step: WaterfallStep, start_df: DataFrame, end_df: DataFrame) -> DataFrame:
    group_by_columns = [step.labelsColumn] + step.groupby
    if step.parentsColumn is not None:
        group_by_columns = group_by_columns + [step.parentsColumn]
    start_df = (
        start_df.groupby(by=group_by_columns)
        .agg({step.valueColumn + '_start': 'sum'})
        .merge(start_df[group_by_columns], on=group_by_columns)
        .rename(columns={step.labelsColumn + '_x': step.labelsColumn})
        .drop_duplicates()
    )

    end_df = (
        end_df.groupby(by=group_by_columns)
        .agg({step.valueColumn + '_end': 'sum'})
        .merge(end_df[group_by_columns], on=group_by_columns)
        .rename(columns={step.labelsColumn + '_x': step.labelsColumn})
        .drop_duplicates()
    )
    # we join the result to compare them
    merged_df = start_df.merge(end_df, on=get_join_key(step))
    merged_df[_RESULT_COLUMN] = (
        merged_df[f'{step.valueColumn}_end'] - merged_df[f'{step.valueColumn}_start']
    )
    merged_df = merged_df.drop(
        columns=[
            f'{step.valueColumn}_start',
            f'{step.valueColumn}_end',
        ]
    )

    # if there is a parent column, we need to aggregate for them
    if step.parentsColumn is not None:
        parents_results = merged_df.groupby(
            step.groupby + [step.parentsColumn], as_index=False
        ).agg({_RESULT_COLUMN: 'sum'})
        parents_results[step.labelsColumn] = parents_results[step.parentsColumn]
        return pd.concat([merged_df, parents_results])
    return merged_df


# the waterfall has a very specific structure.
# this methods create the top / bottom rows.
# these contains the sum of values for the whole milestone, regardless of label.
def compute_agg_milestone(
    step: WaterfallStep, df: DataFrame, start_df: DataFrame, milestone
) -> pd.DataFrame:
    if len(step.groupby) > 0:
        groups = df[step.groupby].drop_duplicates()
        group_by = step.groupby
    else:
        groups = pd.DataFrame({'_VQB_GROUP': [0]})  # pseudo group
        start_df['_VQB_GROUP'] = 0
        groups['_VQB_GROUP'] = 0
        group_by = ['_VQB_GROUP']

    groups = groups.assign(**{step.labelsColumn: milestone})

    agg = start_df.groupby(group_by).agg({f'{step.valueColumn}': 'sum'})
    agg = (
        groups.merge(agg, on=group_by)
        .sort_values(
            by=step.valueColumn if step.sortBy == 'value' else step.labelsColumn,
            ascending=step.order == 'asc',
        )
        .rename(columns={step.labelsColumn: LABEL_WATERFALL_COLUMN})
    )

    if step.parentsColumn is not None:
        agg[GROUP_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
    agg[TYPE_WATERFALL_COLUMN] = None
    agg[LABEL_WATERFALL_COLUMN] = agg[LABEL_WATERFALL_COLUMN].astype(str)
    if len(step.groupby) == 0:
        del start_df['_VQB_GROUP']
        del agg['_VQB_GROUP']
    return agg


# this method shapes the merged DF so it matches the waterfall format
def compute_mid(step: WaterfallStep, merged_df: DataFrame, df: DataFrame) -> DataFrame:
    result_df = DataFrame(
        {
            LABEL_WATERFALL_COLUMN: merged_df.sort_values(
                by=get_sort_column(step), ascending=step.order == 'asc'
            )[step.labelsColumn].astype(str)
        }
    )
    result_df[step.groupby] = merged_df.sort_values(
        by=get_sort_column(step), ascending=step.order == 'asc'
    )[step.groupby]

    if step.parentsColumn is None:
        result_df[TYPE_WATERFALL_COLUMN] = 'Parent'
    else:
        result_df[GROUP_WATERFALL_COLUMN] = merged_df.sort_values(
            by=get_sort_column(step), ascending=step.order == 'asc'
        )[step.parentsColumn]
        result_df[TYPE_WATERFALL_COLUMN] = numpy.where(
            result_df['LABEL_waterfall'] == (result_df[GROUP_WATERFALL_COLUMN]),
            ['parent'],
            ['child'],
        )

    result_df[step.valueColumn] = merged_df.sort_values(
        by=get_sort_column(step), ascending=step.order == 'asc'
    )[_RESULT_COLUMN]
    return result_df


def get_sort_column(step: WaterfallStep):
    if step.sortBy == 'value':
        return _RESULT_COLUMN
    else:
        return step.labelsColumn


def get_join_key(step: WaterfallStep):
    if step.parentsColumn is None:
        return [step.labelsColumn] + step.groupby
    else:
        return [step.labelsColumn, step.parentsColumn] + step.groupby



from typing import List, Optional, Union

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import ToDateStep


def translate_todate(step: ToDateStep) -> List[MongoStep]:
    date_format = step.format

    col_as_string = {'$toString': f'${step.column}'}

    if date_format:
        # % B and %b should be equivalent
        date_format = date_format.replace('%B', '%b')

    if not date_format:
        # Mongo will try to guess the date format
        return [
            {
                '$addFields': {
                    step.column: {
                        '$dateFromString': {
                            'dateString': col_as_string,
                            # However, we give it a second chance to find formats it can't guess natively
                            'onError': {
                                '$cond': [
                                    # Integer values may be either years or timestamps
                                    {'$in': [{'$type': f'${step.column}'}, ['int', 'long']]},
                                    {
                                        '$cond': [
                                            # We decide that values lower than 10_000 will be interpreted as years...
                                            {'$lt': [f'${step.column}', 10_000]},
                                            {
                                                '$dateFromString': {
                                                    'dateString': {
                                                        '$concat': [col_as_string, '-01-01']
                                                    },
                                                    'format': '%Y-%m-%d',
                                                    'onError': {'$literal': None},
                                                }
                                            },
                                            # ...and greater values will be interpreted as timestamps
                                            {
                                                '$convert': {
                                                    'input': f'${step.column}',
                                                    'to': 'date',
                                                    'onError': {'$literal': None},
                                                }
                                            },
                                        ]
                                    },
                                    # otherwise, try to interpret as years
                                    {
                                        '$dateFromString': {
                                            'dateString': {'$concat': [col_as_string, '-01-01']},
                                            'format': '%Y-%m-%d',
                                            # and give up if it doesn't work
                                            'onError': {'$literal': None},
                                        }
                                    },
                                ]
                            },
                        }
                    }
                }
            }
        ]

    # Mongo doesn't handle months names (%b and %B), so we replace them by numbers before converting
    # All these could probably be factorized to make it easier to read!
    elif date_format == '%d %b %Y':
        return [
            {'$addFields': {'_vqbTempArray': {'$split': [col_as_string, ' ']}}},
            _extract_date_parts_to_temp_fields(2, 1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column,
                [
                    '$_vqbTempDay',
                    '-',
                    '$_vqbTempMonth',
                    '-',
                    '$_vqbTempYear',
                ],
                '%d-%m-%Y',
            ),
            _clean_temp_fields(),
        ]

    elif date_format == '%d-%b-%Y':
        return [
            {'$addFields': {'_vqbTempArray': {'$split': [col_as_string, '-']}}},
            _extract_date_parts_to_temp_fields(2, 1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column,
                [
                    '$_vqbTempDay',
                    '-',
                    '$_vqbTempMonth',
                    '-',
                    '$_vqbTempYear',
                ],
                '%d-%m-%Y',
            ),
            _clean_temp_fields(),
        ]

    elif date_format == '%b %Y':
        return [
            {'$addFields': {'_vqbTempArray': {'$split': [col_as_string, ' ']}}},
            _extract_date_parts_to_temp_fields(1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column, ['01-', '$_vqbTempMonth', '-', '$_vqbTempYear'], '%d-%m-%Y'
            ),
            _clean_temp_fields(),
        ]

    elif date_format == '%b-%Y':
        return [
            {'$addFields': {'_vqbTempArray': {'$split': [col_as_string, '-']}}},
            _extract_date_parts_to_temp_fields(1, 0),
            MONTH_REPLACEMENT_STEP,
            _concat_fields_to_date(
                step.column, ['01-', '$_vqbTempMonth', '-', '$_vqbTempYear'], '%d-%m-%Y'
            ),
            _clean_temp_fields(),
        ]

    # Mongo does not support dates where some parts of the date are missing
    elif date_format == '%Y-%m':
        return [
            _concat_fields_to_date(step.column, [col_as_string, '-01'], '%Y-%m-%d'),
        ]

    elif date_format == '%Y/%m':
        return [
            _concat_fields_to_date(step.column, ['/01', col_as_string], '%Y/%m/%d'),
        ]

    elif date_format == '%m-%Y':
        return [
            _concat_fields_to_date(step.column, ['01-', col_as_string], '%d-%m-%Y'),
        ]

    elif date_format == '%m/%Y':
        return [
            _concat_fields_to_date(step.column, ['01/', col_as_string], '%d/%m/%Y'),
        ]

    elif date_format == '%Y':
        return [
            _concat_fields_to_date(step.column, [col_as_string, '-01-01'], '%Y-%m-%d'),
        ]

    else:
        return [
            {
                '$addFields': {
                    step.column: {
                        '$dateFromString': {
                            'dateString': col_as_string,
                            'format': date_format,
                            'onError': {'$literal': None},
                        }
                    }
                }
            }
        ]


MONTH_NUMBER_TO_NAMES = {
    '01': ['jan', 'jan.', 'january', 'janv', 'janv.', 'janvier'],
    '02': ['feb', 'feb.', 'february', 'fév', 'fev', 'févr.', 'fevr.', 'février'],
    '03': ['mar', 'mar.', 'march', 'mars'],
    '04': ['apr', 'apr.', 'april', 'avr', 'avr.', 'avril'],
    '05': ['may', 'mai'],
    '06': ['june', 'jun.', 'june', 'juin'],
    '07': ['jul', 'jul.', 'july', 'juil', 'juil.', 'juillet'],
    '08': ['aug', 'aug.', 'august', 'août', 'aout'],
    '09': ['sep', 'sep.', 'september', 'sept', 'sept.', 'septembre'],
    '10': ['oct', 'oct.', 'october', 'octobre'],
    '11': ['nov', 'nov.', 'november', 'novembre'],
    '12': ['dec', 'dec.', 'december', 'déc', 'déc.', 'décembre'],
}

MONTH_REPLACEMENT_STEP: MongoStep = {
    '$addFields': {
        '_vqbTempMonth'
        '$switch': {
            'branches': [
                {
                    'case': {'$in': month_names},
                    'then': month_number,
                }
                for month_number, month_names in MONTH_NUMBER_TO_NAMES.items()
            ]
        }
    }
}


def _extract_date_parts_to_temp_fields(
    year_position: int, month_position: Optional[int] = None, day_position: Optional[int] = None
) -> MongoStep:
    date_parts_temp_fields: MongoStep = {
        '_vqbTempYear': {'$arrayElemAt': ['$_vqbTempArray', year_position]},
    }

    if month_position is not None:
        date_parts_temp_fields['_vqbTempMonth'] = {
            '$toLower': {'$arrayElemAt': ['$_vqbTempArray', month_position]}
        }

    if day_position is not None:
        date_parts_temp_fields['_vqbTempDay'] = {'$arrayElemAt': ['$_vqbTempArray', day_position]}

    return {
        '$addFields': date_parts_temp_fields,
    }


def _clean_temp_fields():
    return {
        '$project': {'_vqbTempArray': 0, '_vqbTempMonth': 0, '_vqbTempYear': 0, '_vqbTempDate': 0}
    }


def _concat_fields_to_date(target_col: str, fields: List[Union[str, dict]], format: str):
    return {
        '$addFields': {
            target_col: {
                '$dateFromString': {
                    'dateString': {
                        '$concat': fields,
                    },
                    'format': format,
                },
            },
        },
    }



from weaverbird.pipeline.steps.utils.base import BaseStep

from .absolutevalue import AbsoluteValueStep, AbsoluteValueStepWithVariable
from .addmissingdates import AddMissingDatesStep, AddMissingDatesStepWithVariables
from .aggregate import AggregateStep, AggregateStepWithVariables, Aggregation
from .append import AppendStep, AppendStepWithVariable
from .argmax import ArgmaxStep, ArgmaxStepWithVariable
from .argmin import ArgminStep, ArgminStepWithVariable
from .comparetext import CompareTextStep, CompareTextStepWithVariables
from .concatenate import ConcatenateStep, ConcatenateStepWithVariable
from .convert import ConvertStep
from .cumsum import CumSumStep, CumSumStepWithVariable
from .custom import CustomStep
from .customsql import CustomSqlStep, CustomSqlStepWithVariables
from .date_extract import DateExtractStep, DateExtractStepWithVariable
from .delete import DeleteStep
from .dissolve import DissolveStep
from .domain import DomainStep
from .duplicate import DuplicateStep
from .duration import DurationStep, DurationStepWithVariable
from .evolution import EvolutionStep, EvolutionStepWithVariable
from .fillna import FillnaStep, FillnaStepWithVariable
from .filter import FilterStep, FilterStepWithVariables
from .formula import FormulaStep, FormulaStepWithVariable
from .fromdate import FromdateStep
from .hierarchy import HierarchyStep
from .ifthenelse import IfthenelseStep, IfThenElseStepWithVariables
from .join import JoinStep, JoinStepWithVariable
from .lowercase import LowercaseStep
from .moving_average import MovingAverageStep
from .percentage import PercentageStep
from .pivot import PivotStep, PivotStepWithVariable
from .rank import RankStep, RankStepWithVariable
from .rename import RenameStep, RenameStepWithVariable
from .replace import ReplaceStep, ReplaceStepWithVariable
from .rollup import RollupStep, RollupStepWithVariable
from .select import SelectStep
from .simplify import SimplifyStep
from .sort import SortStep
from .split import SplitStep, SplitStepWithVariable
from .statistics import StatisticsStep
from .substring import SubstringStep
from .table import TableStep
from .text import TextStep, TextStepWithVariable
from .todate import ToDateStep
from .top import TopStep, TopStepWithVariables
from .totals import TotalsStep, TotalsStepWithVariable
from .trim import TrimStep
from .uniquegroups import UniqueGroupsStep, UniqueGroupsStepWithVariable
from .unpivot import UnpivotStep, UnpivotStepWithVariable
from .uppercase import UppercaseStep
from .waterfall import WaterfallStep, WaterfallStepWithVariable



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import UnpivotStep


def translate_unpivot(step: UnpivotStep) -> List[MongoStep]:
    # project_cols to be included in Mongo $project steps
    project_cols = {col: f'${col}' for col in step.keep}
    # object_to_array to be included in the first Mongo $project step
    object_to_array = {col: {'$ifNull': [f'${col}', None]} for col in step.unpivot}

    mongo_pipeline: List[dict] = [
        {
            '$project': {**project_cols, '_vqbToUnpivot': {'$objectToArray': object_to_array}},
        },
        {
            '$unwind': '$_vqbToUnpivot',
        },
        {
            '$project': {
                **project_cols,
                step.unpivot_column_name: '$_vqbToUnpivot.k',
                step.value_column_name: '$_vqbToUnpivot.v',
            },
        },
    ]

    if step.dropna:
        mongo_pipeline.append({'$match': {step.value_column_name: {'$ne': None}}})

    return mongo_pipeline



import logging
from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline import Pipeline, steps
from weaverbird.pipeline.steps import AppendStep, DomainStep

logger = logging.getLogger(__name__)


def translate_append(step: AppendStep) -> List[MongoStep]:
    from weaverbird.backends.mongo_translator.mongo_pipeline_translator import translate_pipeline

    pipelines = step.pipelines
    pipelines_names = ['$_vqbPipelineInline']
    lookups = []

    for i, sub_pipeline in enumerate(pipelines):
        pipeline_without_domain = Pipeline(steps=[])

        if isinstance(sub_pipeline, str):
            domain_step = DomainStep(name='domain', domain=sub_pipeline)
        else:
            try:
                assert isinstance(sub_pipeline, list)
            except AssertionError:  # in this case sub_pipeline is a Reference
                raise Exception('References must be resolved before translating the pipeline')

            domain_step = DomainStep(**sub_pipeline[0])
            pipeline_without_domain.steps = [
                getattr(steps, f"{s['name'].capitalize()}Step")(**s) for s in sub_pipeline[1:]
            ]
        lookups.append(
            {
                '$lookup': {
                    'from': domain_step.domain,
                    'pipeline': translate_pipeline(pipeline_without_domain),
                    'as': f'_vqbPipelineToAppend_{i}',
                }
            }
        )
        pipelines_names.append(f'$_vqbPipelineToAppend_{i}')
    return [
        {'$group': {'_id': None, '_vqbPipelineInline': {'$push': '$$ROOT'}}},
        *lookups,
        {'$project': {'_vqbPipelinesUnion': {'$concatArrays': pipelines_names}}},
        {'$unwind': '$_vqbPipelinesUnion'},
        {'$replaceRoot': {'newRoot': '$_vqbPipelinesUnion'}},
    ]



import itertools
from typing import Dict, List, Tuple

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import TotalsStep


def combinations(iterable: list) -> list:
    """combinations([1,2,3]) --> (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"""
    return list(
        itertools.chain.from_iterable(
            itertools.combinations(iterable, r) for r in range(1, len(iterable) + 1)
        )
    )


def column_map(s: List[str]) -> Dict[str, str]:
    return {e: f'${e}' for e in s}


def translate_totals(step: TotalsStep) -> List[MongoStep]:
    facet: Dict[str, List[MongoStep]] = {}
    groups: List[str] = step.groups or []
    add_fields: MongoStep = {}
    project: MongoStep = {'_id': 0}  # ensures $project stage will never be empty
    # list of columns to combine
    to_combine: List[str] = [c.total_column for c in step.total_dimensions]
    # get combinations, remove last combo (most granular combination of columns
    # so not useful to compute total rows) and add an empty tuple (to compute the grand total)
    combos: List[Tuple] = combinations(to_combine)[:-1]
    combos.append(tuple())

    for agg_step in step.aggregations:
        for i in range(len(agg_step.columns)):
            add_fields[agg_step.new_columns[i]] = f'${agg_step.columns[i]}'
            if agg_step.new_columns[i] != agg_step.columns[i]:
                project[agg_step.columns[i]] = 0

    facet['originalData'] = [{'$addFields': add_fields}, {'$project': project}]

    for i in range(len(combos)):
        comb = combos[i]
        # List of columns that that will be used to group the aggregations computation
        # i.e. we will compute total rows for dimensions not included in this group id
        id = column_map(list(comb) + list(groups))
        aggs: Dict[str, dict] = {}
        # get columns not in aggregation, i.e. columns that will hold the total rows labels
        total_columns: List[str] = [e for e in to_combine if e not in comb]
        count_distinct_add_fields = {}

        for agg_step in step.aggregations:
            for j in range(len(agg_step.columns)):
                value_col = agg_step.columns[j]
                aggregated_col = agg_step.new_columns[j]
                agg_func = agg_step.agg_function
                if agg_func == 'count':
                    aggs[aggregated_col] = {'$sum': 1}
                elif agg_func == 'count distinct':
                    # build a set of unique values
                    aggs[aggregated_col] = {'$addToSet': f'${value_col}'}
                    # count the number of items in the set
                    count_distinct_add_fields[aggregated_col] = {'$size': f'${aggregated_col}'}
                else:
                    aggs[aggregated_col] = {f'${agg_func}': f'${value_col}'}

        add_fields_to_add_to_pipeline = (
            [{'$addFields': count_distinct_add_fields}] if count_distinct_add_fields else []
        )

        facet[f'combo_{i}'] = [
            {
                '$group': {
                    '_id': id,
                    **aggs,
                },
            },
            *add_fields_to_add_to_pipeline,
            {
                '$project': {
                    '_id': 0,
                    # Return id fields (untouched)
                    **{k: f'$_id.{k}' for k in id.keys()},
                    # Return computed aggregation fields
                    **{k: 1 for k in aggs.keys()},
                    # Project the label of total rows for those columns
                    **{
                        dimension.total_column: dimension.total_rows_label
                        for dimension in step.total_dimensions
                        if dimension.total_column in total_columns
                    },
                },
            },
        ]

    return [
        {'$facet': facet},
        {
            '$project': {
                '_vqbCombos': {'$concatArrays': [f'${k}' for k in facet]},
            },
        },
        {'$unwind': '$_vqbCombos'},
        {'$replaceRoot': {'newRoot': '$_vqbCombos'}},
    ]



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SelectStep


def translate_select(step: SelectStep) -> List[MongoStep]:
    return [{'$project': {f'{column}': 1 for column in step.columns}}]



from typing import TypeVar

from pypika.dialects import SnowflakeQuery
from pypika.terms import CustomFunction, Field, LiteralValue, Term

from weaverbird.backends.pypika_translator.dialects import SQLDialect
from weaverbird.backends.pypika_translator.operators import FromDateOp, RegexOp, ToDateOp
from weaverbird.backends.pypika_translator.translators.base import DataTypeMapping, SQLTranslator
from weaverbird.pipeline.steps.date_extract import DATE_INFO

Self = TypeVar("Self", bound="SQLTranslator")


class SnowflakeTranslator(SQLTranslator):
    DIALECT = SQLDialect.SNOWFLAKE
    QUERY_CLS = SnowflakeQuery
    DATA_TYPE_MAPPING = DataTypeMapping(
        boolean="BOOLEAN",
        date="DATE",
        float="DOUBLE PRECISION",
        integer="INTEGER",
        text="TEXT",
        datetime="TIMESTAMP",
        timestamp="TIMESTAMP",
    )
    SUPPORT_ROW_NUMBER = True
    SUPPORT_SPLIT_PART = True
    SUPPORT_UNPIVOT = True
    FROM_DATE_OP = FromDateOp.TO_CHAR
    REGEXP_OP = RegexOp.REGEXP
    TO_DATE_OP = ToDateOp.TO_TIMESTAMP_NTZ

    @classmethod
    def _add_date(
        cls, *, date_column: Field, add_date_value: int, add_date_unit: DATE_INFO
    ) -> Term:
        add_date_func = CustomFunction('DATEADD', ['interval', 'increment', 'datecol'])
        return add_date_func(add_date_unit, add_date_value, date_column)

    @classmethod
    def _get_date_extract_func(cls, *, date_unit: DATE_INFO, target_column: Field) -> Term:
        if date_unit.lower() in [
            "seconds",
            "minutes",
            "hour",
            "day",
            "dayofweekiso",
            "dayofyear",
            "week",
            "weekiso",
            "month",
            "quarter",
            "year",
            "yearofweek",
            "yearofweekiso",
        ]:
            return LiteralValue(
                f"EXTRACT({date_unit[:-1] if date_unit in ['seconds', 'minutes'] else date_unit} "
                f"from to_timestamp({target_column.name}))"
            )
        else:
            date_func: dict[DATE_INFO, str] = {
                # Returning numbers
                # -----------------
                "milliseconds": "ROUND(EXTRACT(nanosecond FROM to_timestamp(____target____))/1000000)",
                "isoDayOfWeek": "DAYOFWEEKISO(to_timestamp(____target____))",
                "isoWeek": "WEEKISO(to_timestamp(____target____))",
                "isoYear": "YEAROFWEEKISO(to_timestamp(____target____))",
                # same problem as 'week' behaviour
                "previousWeek": "WEEK(to_timestamp(____target____) - interval '1 week')",
                "previousIsoWeek": "WEEKISO(to_timestamp(____target____) - interval '1 week')",
                "previousMonth": "MONTH(to_timestamp(____target____) - interval '1 month')",
                "previousQuarter": "QUARTER(to_timestamp(____target____) - interval '1 quarter')",
                "previousYear": "YEAR(to_timestamp(____target____) - interval '1 year')",
                "dayOfWeek": "MOD(DAYOFWEEKISO(to_timestamp(____target____)), 7) + 1",
                # Returning dates
                # ---------------
                "previousDay": "DATE_TRUNC(day, to_timestamp(____target____) - interval '1 day')",
                "firstDayOfWeek": "DATE_TRUNC(day, DATEADD(day, -(MOD(DAYOFWEEKISO(to_timestamp(____target____)), 7) + 1)+1, to_timestamp(____target____)))",
                "firstDayOfIsoWeek": "DATE_TRUNC(day, DATEADD(day, -DAYOFWEEKISO(to_timestamp(____target____))+1, to_timestamp(____target____)))",
                "firstDayOfMonth": "TO_TIMESTAMP_NTZ(DATE_TRUNC(month, to_timestamp(____target____)))",
                "firstDayOfQuarter": "TO_TIMESTAMP_NTZ(DATE_TRUNC(quarter, to_timestamp(____target____)))",
                "firstDayOfYear": "TO_TIMESTAMP_NTZ(DATE_TRUNC(year, to_timestamp(____target____)))",
                "firstDayOfPreviousWeek": "DATE_TRUNC(day, DATEADD(day, -(MOD(DAYOFWEEKISO(to_timestamp(____target____)), 7) + 1)+1, to_timestamp(____target____))) - interval '1 week'",
                "firstDayOfPreviousIsoWeek": "DATE_TRUNC(day, DATEADD(day, -DAYOFWEEKISO(to_timestamp(____target____))+1, to_timestamp(____target____))) - interval '1 week'",
                "firstDayOfPreviousMonth": "TO_TIMESTAMP_NTZ(DATE_TRUNC(month, to_timestamp(____target____))) - interval '1 month'",
                "firstDayOfPreviousQuarter": "TO_TIMESTAMP_NTZ(DATE_TRUNC(quarter, to_timestamp(____target____))) - interval '1 quarter'",
                "firstDayOfPreviousYear": "TO_TIMESTAMP_NTZ(DATE_TRUNC(year, to_timestamp(____target____))) - interval '1 year'",
            }
            return LiteralValue(
                f'{date_func[date_unit].replace("____target____", target_column.name)}'
            )


SQLTranslator.register(SnowflakeTranslator)



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps.moving_average import MovingAverageStep


def translate_moving_average(step: MovingAverageStep) -> List[MongoStep]:
    return [
        # Ensure the reference column is sorted to prepare for the moving average computation
        {'$sort': {step.column_to_sort: 1}},
        # If needing to group the computation, provide for the columns, else _id is set to None
        {
            '$group': {
                '_id': {col: f'${col}' for col in step.groups} if step.groups else None,
                '_vqbArray': {'$push': '$$ROOT'},
            },
        },
        # Prepare an array of documents with a new moving average field. One array per group (if any)
        {
            '$addFields': {
                '_vqbArray': {
                    # We use $map to apply operations while looping over the _vqbArray documents
                    '$map': {
                        # We create an index ("idx") variable that will go from 0 to the size of _vqbArray
                        'input': {'$range': [0, {'$size': '$_vqbArray'}]},
                        'as': 'idx',
                        # We will use the idx variable in the following stages
                        'in': {
                            '$cond': [
                                # If the index is less than the moving window minus 1...
                                {
                                    '$lt': ['$$idx', (step.moving_window) - 1]
                                },  # explicit type for typescript
                                # ... then we cannot apply the moving average computation, and
                                # we just keep the original document without any new field...
                                {'$arrayElemAt': ['$_vqbArray', '$$idx']},
                                # ... else we compute the value average over the last N documents starting
                                # from the current index document, N being equal to the moving window)
                                {
                                    '$mergeObjects': [
                                        # Keep track of original document
                                        {'$arrayElemAt': ['$_vqbArray', '$$idx']},
                                        # and add the new moving average column
                                        {
                                            step.new_column_name
                                            or f'{step.value_column}_MOVING_AVG': {
                                                '$avg': {
                                                    '$slice': [
                                                        f'$_vqbArray.{step.value_column}',
                                                        {
                                                            '$subtract': [
                                                                '$$idx',
                                                                step.moving_window - 1,
                                                            ]
                                                        },  # explicit type for typescript
                                                        step.moving_window,
                                                    ],
                                                },
                                            },
                                        },
                                    ],
                                },
                            ],
                        },
                    },
                },
            },
        },
        # Flatten the array(s) to get back to 1 row per document (at this stage in the field "_vqbArray")
        {'$unwind': '$_vqbArray'},
        # Set the _vqbArray field as the new root to get back to the original document granularity
        {'$replaceRoot': {'newRoot': '$_vqbArray'}},
    ]



from distutils import log

from weaverbird.backends.sql_translator.steps.utils.query_transformation import (
    build_selection_query,
)
from weaverbird.backends.sql_translator.types import (
    SQLDialect,
    SQLPipelineTranslator,
    SQLQuery,
    SQLQueryDescriber,
    SQLQueryExecutor,
    SQLQueryRetriever,
)
from weaverbird.pipeline.steps import RankStep


def translate_rank(
    step: RankStep,
    query: SQLQuery,
    index: int,
    sql_query_retriever: SQLQueryRetriever = None,
    sql_query_describer: SQLQueryDescriber = None,
    sql_query_executor: SQLQueryExecutor = None,
    sql_translate_pipeline: SQLPipelineTranslator = None,
    subcall_from_other_pipeline_count: int = None,
    sql_dialect: SQLDialect = None,
) -> SQLQuery:
    query_name = f'RANK_STEP_{index}'

    log.debug(
        '############################################################'
        f'query_name: {query_name}\n'
        '------------------------------------------------------------'
        f'step: {step}\n'
        f'query.transformed_query: {query.transformed_query}\n'
        f'query.metadata_manager.query_metadata: {query.metadata_manager.retrieve_query_metadata()}\n'
    )

    rank_mode = "DENSE_RANK()" if step.method == "dense" else "RANK()"
    step.new_column_name = (
        f"{step.value_col}_RANK" if step.new_column_name is None else step.new_column_name
    )

    # the rank query
    rank_query: str = ""
    order_by_query: str = ""
    if len(step.groupby) > 0:
        rank_query = (
            f", ({rank_mode} OVER (PARTITION BY {', '.join(step.groupby)} "
            f"ORDER BY {step.value_col} {step.order})) AS {step.new_column_name}"
        )

        order_on_groupby = ' ASC, '.join(step.groupby)
        if len(step.groupby) > 0:
            order_on_groupby += ' ASC'

        order_by_query = f"ORDER BY {step.new_column_name} ASC, {order_on_groupby}"
    else:
        rank_query = (
            f", ({rank_mode} OVER ("
            f"ORDER BY {step.value_col} {step.order})) AS {step.new_column_name}"
        )
        order_by_query = f"ORDER BY {step.new_column_name} ASC"

    final_query = (
        f" (SELECT {query.metadata_manager.retrieve_query_metadata_columns_as_str()}"
        f"{rank_query}"
        f" FROM {query.query_name} {order_by_query})"
    )

    # we add the column to the metadata
    query.metadata_manager.add_query_metadata_column(step.new_column_name, 'int')

    new_query = SQLQuery(
        query_name=query_name,
        transformed_query=f"{query.transformed_query}, {query_name} AS {final_query}",
        selection_query=build_selection_query(
            query.metadata_manager.retrieve_query_metadata_columns(), query_name
        ),
        metadata_manager=query.metadata_manager,
    )

    log.debug(
        '------------------------------------------------------------'
        f'SQLquery: {new_query.transformed_query}'
        '############################################################'
    )

    return new_query



from typing import List

from weaverbird.backends.mongo_translator.steps.types import MongoStep
from weaverbird.pipeline.steps import SortStep


def translate_sort(step: SortStep) -> List[MongoStep]:
    return [
        {
            '$sort': {
                sort_column.column: 1 if sort_column.order == 'asc' else -1
                for sort_column in step.columns
            }
        }
    ]



from typing import List, Literal, Optional

from pydantic import Field

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.types import ColumnName


class MovingAverageStep(BaseStep):
    name: Literal['movingaverage'] = 'movingaverage'

    value_column: Optional[ColumnName] = Field(None, alias='valueColumn')
    column_to_sort: ColumnName = Field(alias='columnToSort')
    moving_window: int = Field(alias='movingWindow')
    groups: List[ColumnName] = []
    new_column_name: Optional[ColumnName] = Field(alias='newColumnName')



from pandas import DataFrame

from weaverbird.backends.pandas_executor.steps.utils.formula import eval_formula
from weaverbird.backends.pandas_executor.types import DomainRetriever, PipelineExecutor
from weaverbird.pipeline.steps import FormulaStep


def execute_formula(
    step: FormulaStep,
    df: DataFrame,
    domain_retriever: DomainRetriever = None,
    execute_pipeline: PipelineExecutor = None,
) -> DataFrame:
    serie = eval_formula(df, step.formula)
    return df.assign(**{step.new_column: serie})



from typing import List, Literal, Tuple, Union

from pydantic import Field, root_validator

from weaverbird.pipeline.steps.utils.base import BaseStep
from weaverbird.pipeline.steps.utils.render_variables import StepWithVariablesMixin
from weaverbird.pipeline.types import TemplatedVariable


class RenameStep(BaseStep):
    name: Literal['rename'] = 'rename'
    to_rename: List[Tuple[str, str]] = Field(..., alias='toRename')

    @root_validator(pre=True)
    def handle_legacy_syntax(cls, values):
        if 'oldname' in values and 'newname' in values:
            values['to_rename'] = [(values.pop('oldname'), values.pop('newname'))]
        return values


class RenameStepWithVariable(RenameStep, StepWithVariablesMixin):
    to_rename: Union[TemplatedVariable, List[Tuple[TemplatedVariable, TemplatedVariable]]] = Field(
        ..., alias='toRename'
    )
